unit feeds;

interface

uses
  ShellApi, Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Inifiles, Dialogs, StdCtrls,IdHTTP,LibXmlParser,utils, ExtCtrls, ComCtrls,pagefound,
  shortintmap,compmap,splashscreen,strutils;

type
   relType = shortint;
   T2DRelTypeArray = array of array of relType;

   // represents a news item found on a feed
   TItem = Record
      title: string;
      description: string;
      link: string;
      header: string;
      relevanceRatingFromHistory: integer;
      relevanceRatingFromUser: integer;
   end;

   // represents the details of a particular feed
   TFeed = Record
      url: string;
      active: boolean;
      keywords : TStringList;
      refreshInterval: integer;
      lastRefresh: COMP;
   end;

   Items = array of TItem;

   // a link between two words
   TWordLink = Record
      word1: integer;
      word2: integer;
      strength: relType;
   End;

   TWordLinks = array of TWordLink;


  TfrmFeeds = class(TForm)
    Panel1: TPanel;
    txtResults: TMemo;
    Label4: TLabel;
    Panel2: TPanel;
    Panel4: TPanel;
    Panel5: TPanel;
    lblAddFeedURL: TLabel;
    txtAddFeedURL: TEdit;
    btnAddFeed: TButton;
    Timer1: TTimer;
    Splitter1: TSplitter;
    Panel3: TPanel;
    Panel7: TPanel;
    lblRefreshInterval: TLabel;
    cbActive: TCheckBox;
    txtRefreshInterval: TEdit;
    Panel8: TPanel;
    Label2: TLabel;
    txtKeywords: TMemo;
    btnUpdate: TButton;
    Panel9: TPanel;
    Feeds: TLabel;
    lbFeeds: TListBox;
    btnUnitTests: TButton;
    btnAllInactive: TButton;
    btnAllActive: TButton;
    btnDelete: TButton;
    btnSearchAllFeeds: TButton;
    btnLearn: TButton;
    btnReset: TButton;
    btnOpenFeedInBrowser: TButton;
    btnSearchSelectedFeed: TButton;
    lblShowRelevance: TLabel;
    tbShowRelevance: TTrackBar;
    Label3: TLabel;
    rgResultsFormat: TRadioGroup;
    btnCancel: TButton;
    btnCheckRel: TButton;
    cbCommonIntervals: TComboBox;
    procedure btnAddFeedClick(Sender: TObject);
    procedure btnSearchAllFeedsClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure lbFeedsClick(Sender: TObject);
    procedure cbActiveClick(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure txtRefreshIntervalChange(Sender: TObject);
    procedure txtAddFeedURLChange(Sender: TObject);
    procedure btnDeleteClick(Sender: TObject);
    procedure txtKeywordsExit(Sender: TObject);
    procedure btnUpdateClick(Sender: TObject);
    procedure txtKeywordsChange(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnAllActiveClick(Sender: TObject);
    procedure btnAllInactiveClick(Sender: TObject);
    procedure btnResetClick(Sender: TObject);
    procedure btnOpenFeedInBrowserClick(Sender: TObject);
    procedure btnSearchSelectedFeedClick(Sender: TObject);
    procedure btnLearnClick(Sender: TObject);
    procedure btnUnitTestsClick(Sender: TObject);
    procedure btnCancelClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure btnCheckRelClick(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure rgResultsFormatClick(Sender: TObject);
    procedure tbShowRelevanceChange(Sender: TObject);
    procedure txtRefreshIntervalExit(Sender: TObject);
    procedure txtKeywordsClick(Sender: TObject);
    procedure txtAddFeedURLClick(Sender: TObject);
    procedure txtRefreshIntervalEnter(Sender: TObject);
    procedure cbCommonIntervalsChange(Sender: TObject);
    procedure cbCommonIntervalsExit(Sender: TObject);
  private
    { Private declarations }
    wordLinks: TWordLinks;
    frmOwner: TForm;
    lastSavedConfig, saveConfigInterval, lastUserInteraction: COMP;
    minWordLength,maxWordLength,maxWordWeight: integer;
    commonWords: TStringList;
    relevantWords: TShortIntMap;
    urlsShownRecently: TCompMap;
    learning,shownHelp: boolean;
    procedure sortFeeds;
    procedure scanFeed(var feed: TFeed; var listToAddTo:Items);
    function scanKeywords(var feed: TFeed; textToSearch: string): string;
    function downloadFile(url: string): string;
    procedure updateFeedDetails;
    procedure enableGui(enabled: boolean);
    procedure log(line: string; isVerbose: boolean);
    procedure addUrlToShownRecentlyList(url: string);
    function itemInList(var item:TItem; var itemList: Items): boolean;
    function urlShownRecently(url: string): boolean;
    procedure setAllActive(active: boolean);
    procedure doScan(var feed: TFeed; var item:TItem; var itemList: Items);
    function findLinkIndex(index1, index2: integer): integer;
    procedure updateRelevance(relevance: integer; theText: string);
    procedure updateRelTable(var words: TStringList; relevance: integer);
    procedure handleFoundItems(var matchingItems: Items);
    procedure sortItems(var itemsToSort: Items);
    procedure showFoundItemsAsTable(var matchingItems: Items);
    procedure showFoundItemsOneByOne(var matchingItems: Items);
    procedure loadWordWeightingsTable(filename: string);
    function loadOldRelTable(filename: string): boolean;
    procedure loadLinksTable;
    procedure saveConfig;  overload;
    procedure saveLinksTable;
    procedure cleanData;
    // procedure Save2DShortArray(const A: T2DRelTypeArray; filename: string);
    procedure Load2DShortArray(var A: T2DRelTypeArray; filename: string);
    procedure dumpLinkTable;
  public
    { Public declarations }
    cancelled: boolean;
    procedure setOwner(_frmOwner: TForm);
    procedure loadConfig(confFile: TIniFile);
    procedure saveConfig(confFile: TIniFile);  overload;
    procedure saveData;
    procedure searchAll(startedManually: boolean);
    function checkRelevance(theText: string): integer;
    function checkSimpleRelevance(var words: TStringList): integer;
    function checkRelevanceInRelTable(var words: TStringList): integer;
    function getCleanWordList(theText: string): TStringList;
    function getRelevantWordCount: integer;
  end;



const
   defaultRefreshInterval = 300;
   tempFile = 'temp.xml';
   iniPrefix = 'rss';
   sep = '---***---';
   shownUrlsFile = 'shownfeedurls.dat';
   wordWeightingsFile = 'wordweightings.dat';
   oldRelTableFile = 'reltable.dat';
   linksTableFile = 'wordlinks.dat';

var
  frmFeeds: TfrmFeeds;
  downloadDir,dataDir: string;
  feedsList: array of TFeed;
  updatingGUI: boolean;

       


implementation

uses main,multiplepagesfound;


{$R *.dfm}


procedure TfrmFeeds.setOwner(_frmOwner: TForm);
begin
   self.frmOwner := _frmOwner;
end;

procedure TfrmFeeds.saveConfig;
var
   confFile: TIniFile;
begin
   cbCommonIntervals.Visible := false;
   confFile := TIniFile.Create(ExtractFilePath(Application.ExeName) + '\' + confFileName);
   try
      saveConfig(confFile);
   finally
      confFile.free;
   end;
end;

procedure TfrmFeeds.btnAddFeedClick(Sender: TObject);
begin
   if uppercase(copy(txtAddFeedURL.Text, 1, 7)) <> 'HTTP://' then
   begin
      ShowMessage('Feed URL must start with HTTP://');
      txtAddFeedURL.SetFocus;
      exit;
   end;
   lastUserInteraction := CurrentTimeMillis;
   lbFeeds.items.add(txtAddFeedURL.Text);
   setLength(feedsList, length(feedsList)+1);
   with feedsList[length(feedsList)-1] do
   begin
      url := txtAddFeedURL.Text;
      active := cbActive.Checked;
      keywords := parseString(txtKeywords.Text,chr(13)+chr(10));
      try
         refreshInterval := StrToInt(txtRefreshInterval.text);
      except
         refreshInterval := defaultRefreshInterval;
      end;
   end;
   lbFeeds.ItemIndex := lbFeeds.Count-1;
   sortFeeds;
   log('Added new feed',false);
   SaveConfig;
end;

procedure TfrmFeeds.sortFeeds;
var
   i,j: integer;
   tmpFeed: TFeed;
begin
   for i := 0 to length(feedsList)-1 do
   begin
      for j := 0 to length(feedsList)-1 do
      begin
         if feedsList[i].url > feedsList[j].url then
         begin
            tmpFeed := feedsList[i];
            feedsList[i] := feedsList[j];
            lbFeeds.items[i] := feedsList[j].url;
            feedsList[j] := tmpFeed;
            lbFeeds.items[j] := tmpFeed.url;
         end;
      end;
   end;
end;

procedure TfrmFeeds.btnAllActiveClick(Sender: TObject);
begin
   setAllActive(true);
end;

procedure TfrmFeeds.btnAllInactiveClick(Sender: TObject);
begin
   setAllActive(false);
end;


procedure TfrmFeeds.btnCancelClick(Sender: TObject);
begin
   cancelled := true;
end;

procedure TfrmFeeds.setAllActive(active: boolean);
var
   i: integer;
begin
   lastUserInteraction := CurrentTimeMillis;
   for i := 0 to lbFeeds.Count-1 do
      feedsList[i].active := active;

   cbActive.Checked := active;
   SaveConfig;
end;

procedure TfrmFeeds.btnDeleteClick(Sender: TObject);
var
   i: integer;
begin
   timer1.Enabled := false;
   try
      if lbFeeds.ItemIndex > -1 then
      begin
         if MessageDlg('Delete feed?', mtConfirmation, [mbYes, mbNo], 0) = mrYes then
         begin
            for i := lbFeeds.ItemIndex to lbFeeds.Count-2 do
               feedsList[i] := feedsList[i+1];
            setLength(feedsList, length(feedsList)-1);
            log('Deleted feed ' + IntToStr(lbFeeds.ItemIndex),false);
            lbFeeds.Items.Delete(lbFeeds.ItemIndex);
            SaveConfig;
         end;
      end
      else
         ShowMessage('Please select a feed to delete');
   finally
      timer1.Enabled := true;
   end;
end;

procedure TfrmFeeds.btnLearnClick(Sender: TObject);
var
   feed: TFeed;
   copyUrlsShownRecently: TCompMap;
   matchingItems: Items;
begin
   cancelled := false;
   timer1.Enabled := false;
   copyUrlsShownRecently := TCompMap.Create;

   try
      if lbFeeds.ItemIndex > -1 then
      begin
         if firstTime then ShowMessage('I will now show you all the articles from this feed.' +chr(13)+chr(10)+chr(13)+chr(10) +
            'Please let me know how relevant each article is by using the scale on the form, then I can start to understand the kind of content that interests you.' +chr(13)+chr(10) +chr(13)+chr(10) +
            'The more accurate feedback you give me, the more relevant content I can find for you.');

         learning := true;

         // save urls shown list, then clear it
         copyUrlsShownRecently.AddAll(urlsShownRecently);
         urlsShownRecently.Clear;

         // start learning
         feed := feedsList[lbFeeds.ItemIndex];
         feed.keywords := parseString('1:a,1:e,1:i,1:o,1:u,1:y',',');
         setLength(matchingItems, 0);
         try
            scanFeed(feed,matchingItems);
            handleFoundItems(matchingItems);
         finally
            setLength(matchingItems,0);
         end;

         // reset shown urls list
         urlsShownRecently.Clear;
         urlsShownRecently.AddAll(copyUrlsShownRecently);

         saveConfig;
         saveData;

         ShowMessage('Learning finished for this feed.'+chr(13)+chr(10) +
            'You can repeat this process at any time with this or other feeds.'+chr(13)+chr(10) +
            'The more I learn, the more relevant the articles I show you will be.');
      end
      else
      begin
         ShowMessage('Please select a feed to use for learning');
      end;
   finally
      learning := false;
      timer1.Enabled := true;
      copyUrlsShownRecently.free;
   end;
end;

procedure TfrmFeeds.btnOpenFeedInBrowserClick(Sender: TObject);
begin
   lastUserInteraction := CurrentTimeMillis;
   if lbFeeds.ItemIndex > -1 then
      ShellExecute(GetDesktopWindow, 'open', PChar(feedsList[lbFeeds.ItemIndex].url), nil, nil, SW_ShowNormal)
   else
      ShowMessage('Please select a feed to open');
end;

procedure TfrmFeeds.btnResetClick(Sender: TObject);
var
   i: integer;
begin
   timer1.Enabled := false;
   try
      if MessageDlg('Reset - are you sure?', mtConfirmation, [mbYes, mbNo], 0) = mrYes then
      begin
         urlsShownRecently.Clear;
         for i := 0 to length(feedsList)-1 do
            feedsList[i].lastRefresh := 0;
      end;
   finally
      timer1.Enabled := true;
   end;
end;

procedure TfrmFeeds.btnSearchAllFeedsClick(Sender: TObject);
var
   i: integer;
begin
   for i := 0 to length(feedsList)-1 do
   begin
      if feedsList[i].active then
         feedsList[i].lastRefresh := 0;
   end;

   searchAll(true);
end;


procedure TfrmFeeds.btnSearchSelectedFeedClick(Sender: TObject);
var
   matchingItems: Items;
begin
   cancelled := false;
   timer1.Enabled := false;
   enableGui(false);
   try
      if lbFeeds.ItemIndex > -1 then
      begin
         setLength(matchingItems, 0);
         try
            scanFeed(feedsList[lbFeeds.ItemIndex],matchingItems);
            handleFoundItems(matchingItems);
         finally
            setLength(matchingItems,0);
         end;
      end
      else
         ShowMessage('Please select a feed to search');
   finally
      timer1.Enabled := true;
      enableGui(true);
   end;
end;

procedure TfrmFeeds.saveLinksTable;
var
   f: textfile;
   i: integer;
begin
   AssignFile(f, dataDir + '\' + linksTableFile);
   rewrite(f);
   for i := 0 to length(wordLinks)-1 do
   begin
      frmSplashScreen.updateDetails('Saving ' + linksTableFile,i, length(wordLinks), 10);
      if wordLinks[i].strength <> 0 then
         writeln(f, IntToStr(wordLinks[i].word1) + ',' + IntToSTr(wordLinks[i].word2) + '=' + IntToStr(wordLinks[i].strength));
   end;
   closefile(f);
   frmSplashScreen.Hide;
end;

(*
procedure TfrmFeeds.Save2DShortArray(const A: T2DRelTypeArray; filename: string);
var
  writer: TWriter;
  i: Integer;
  fs: TFileStream;
begin
   fs := TFileStream.create(filename,fmCreate);
   try
     writer := TWriter.Create(fs, 8096);
     try
       writer.WriteInteger(Length(A));
       for i := 0 to Length(A) - 1 do
       begin
         writer.WriteInteger(Length(A[i]));
         writer.Write(A[i, 0], Length(A[i]) * sizeof(A[i, 0]));
       end;
     finally
       writer.Free;
     end;
    finally
      fs.free;
   end;
end;
*)

{purely provided to load and convert the legacy word relationship table}
procedure TfrmFeeds.Load2DShortArray(var A: T2DRelTypeArray; filename: string);
var
  reader: TReader;
  i, numrows, numcols: Integer;
  fs: TFileStream;
begin
   fs := TFileStream.create(filename,fmOpenRead);
   try
     reader := TReader.Create(fs, 8096);
     try
       numrows := reader.ReadInteger;
       SetLength(A, numrows);
       for i := 0 to numrows - 1 do
       begin
         frmSplashScreen.updateDetails('Loading ' + ExtractFileName(filename),i,numrows,10);
         numcols := reader.ReadInteger;
         SetLength(A[i], numcols);
         reader.Read(A[i, 0], numcols * sizeof(A[i, 0]));
       end;
     finally
       reader.Free;
     end;
   finally
      fs.free;
      frmSplashScreen.updateDetails('Loading ' + ExtractFileName(filename),100);
      frmSplashScreen.Hide;
   end;
end;

procedure TfrmFeeds.btnUnitTestsClick(Sender: TObject);
var
   i: integer;
   wordList,cleanWords, sorted: TStringList;
   feed: TFeed;
   sim: TShortIntMap;
begin
   timer1.Enabled := false;
   verboseOutput := true;

   // dumpLinkTable;
   // exit;

   try
      sim := TShortIntMap.create;
      sim.put('EEE',5);
      sim.put('AAA',1);
      sim.put('CCC',3);
      sim.put('BBB',2);
      sim.put('DDD',4);
      assert(sim.get('EEE') = 5);
      assert(sim.get('AAA') = 1);
      assert(sim.get('CCC') = 3);
      assert(sim.get('DDD') = 4);
      assert(sim.get('BBB') = 2);
      sim.free;

      assert(removeExcessSpaces('  test test    test     test                 test  ') = 'test test test test test');
      assert(removeExcessSpaces('test test') = 'test test');
      assert(removeExcessSpaces('test') = 'test');

      cleanWords := getCleanWordList('   and  here''s to john''s apples for the monsters'' party  ');
      // showmessage(cleanWords.getText);
      assert(cleanWords.getText = 'JOHN'+chr(13)+chr(10)+'APPLE'+chr(13)+chr(10)+'MONSTER'+chr(13)+chr(10)+'PARTY'+chr(13)+chr(10));
      cleanWords.free;

      cleanWords := getCleanWordList('in john''s $$$ "report" he said that if "iran''s nuclear programme doesn''t stop soon george bush will get ''really'' cross"!!!, so thEn everyone everyone everyone reads it john''s report report!');
      removeDuplicates(cleanWords);
      // showmessage(cleanWords.getText);
      assert(cleanWords.getText = 'BUSH'+CHR(13)+CHR(10)+'CROSS'+CHR(13)+CHR(10)+'EVERYONE'+CHR(13)+CHR(10)+'GEORGE'+CHR(13)+CHR(10)+'GET'+CHR(13)+CHR(10)+'IRAN'+CHR(13)+CHR(10)+'JOHN'+CHR(13)+CHR(10)+'NUCLEAR'+CHR(13)+CHR(10)+'PROGRAMME'+CHR(13)+CHR(10)+'READ'+CHR(13)+CHR(10)+'REALLY'+CHR(13)+CHR(10)+'REPORT'+CHR(13)+CHR(10)+'SOON'+CHR(13)+CHR(10)+'STOP'+CHR(13)+CHR(10));
      cleanWords.free;

      sorted := TStringList.create;
      sorted.Sorted := true;
      sorted.CaseSensitive := false;
      sorted.add('blah');
      sorted.add('test');
      sorted.add('aaa');
      assert(findStringInSortedList('test',sorted));
      assert(findStringInSortedList('AAA',sorted));
      assert(findStringInSortedList('blah',sorted));
      assert(not findStringInSortedList('xblah',sorted));
      sorted.free;

      assert(removeHtml('bar><nobar>') = '');
      assert(removeHtml('<nobr') = '');
      assert(removeHtml('nobr>') = '');
      assert(removeHtml('<<nobr') = '');
      assert(removeHtml('nobr><blah') = '');
      assert(removeHtml('test<br><p><script language=javascript>function x():</script>end&quot;') = 'test  end"');

      // scan for keyword tests
      feed.keywords := parseString('1:gary mckinnon'+chr(13)+chr(10)+'1:richard,dawkins',chr(13)+chr(10));;
      assert(scanKeywords(feed,'gary is a hacker') = '');
      assert(scanKeywords(feed,'gary mckinnon is a hacker') <> '');
      assert(scanKeywords(feed,'richard dawkins rocks') <> '');
      feed.keywords.free;

      relevantWords.clear;
      setLength(wordLinks, 0);

      // positive test (max weight)
      for i := 1 to 100 do
         updateRelevance(10,'jordan abdullah katie sand');
      dumpLinkTable;
      wordList := getCleanWordList('jordan katie jordan abdullah katie sand');
      assert(checkRelevanceInRelTable(wordList) = 100);
      wordList.free;

      // negative test (max weight)
      for i := 1 to 100 do
         updateRelevance(-10,'jordan katie');
      dumpLinkTable;
      wordList := getCleanWordList('jordan katie katie');
      assert(checkRelevanceInRelTable(wordList) = -100);
      wordList.free;

      // single word test
      wordList := getCleanWordList('jordan');
      assert(checkRelevanceInRelTable(wordList) = 0);
      wordList.free;

      // exclude common words
      wordList := getCleanWordList('jordan AND THE jordan');
      assert(checkRelevanceInRelTable(wordList) = 0);
      wordList.free;

      relevantWords.clear;
      setLength(wordLinks, 0);

      // deep word scan tests
      for i := 1 to 2 do
      begin
         updateRelevance(9,'jordan abdullah');
         updateRelevance(8,'jordan sand');
         updateRelevance(-8,'jordan''s katie');
         updateRelevance(-5,'katie sands');
         updateRelevance(-3,'katie abdullah');
         updateRelevance(7,'sand abdullah''s');
      end;

      dumpLinkTable;

      assert(checkRelevance('jordan abdullah') = 40);
      assert(checkRelevance('abdullahs jordan') = 40);
      assert(checkRelevance('jordan sands') = (((36+40) div 2) + 32) div 2);
      assert(checkRelevance('sand jordan') = (((36+40) div 2) + 32) div 2);
      assert(checkRelevance('jordan katie') = -14);
      assert(checkRelevance('katies jordan') = -14);
      assert(checkRelevance('jordan katie sand abdullah') = 10);
      assert(checkRelevance('abdullah katie''s sand jordan') = 10);

      // FIXME save new links table & load it back
      
      // test save
      {Save2DShortArray(relTable, 'c:\temp\reltable.dat');
      relevantWords.saveToFile('c:\temp\rel.dat');
      relevantWords.clear;
      setLength(relTable, 0, 0);}

      // test load
      {relevantWords.loadFromFile('c:\temp\rel.dat');
      Load2DShortArray(relTable, 'c:\temp\reltable.dat');
      DeleteFile('c:\temp\rel.dat');
      DeleteFile('c:\temp\reltable.dat');}

      // retest data
      cleanData;
      
      assert(checkRelevance('jordan abdullah') = 40);
      assert(checkRelevance('abdullah jordan') = 40);
      assert(checkRelevance('jordans sand') = (((36+40) div 2) + 32) div 2);
      assert(checkRelevance('sand jordan') = (((36+40) div 2) + 32) div 2);
      assert(checkRelevance('jordan katie') = -14);
      assert(checkRelevance('katie''s jordan') = -14);
      assert(checkRelevance('jordan katie sand abdullah') = 10);
      assert(checkRelevance('abdullah katie sand jordan') = 10);

      // long sentence tests
      relevantWords.clear;
      setLength(wordLinks, 0);

      for i := 1 to 5 do
         updateRelevance(10,'the quick brown fox jumped over the big fat bloated lazy dog ');

      wordList := getCleanWordList('quick brown fox jumped over big fat bloated lazy dog');
      assert(checkRelevanceInRelTable(wordList) = 100);
      wordList.free;

      {wordList := getCleanWordList('quick brown wolf jumped over big fat bloated lazy cat');
      assert(checkRelevanceInRelTable(wordList) = 62);
      wordList.free;}

      showmessage('All tests passed');

   finally
      application.terminate; // otherwise we will overwrite the data when we exit
   end;

end;


procedure TfrmFeeds.dumpLinkTable;
var
   f: textfile;
   i: integer;
begin
   AssignFile(f, 'c:\temp\dump.txt');
   rewrite(f);
   for i := 0 to length(wordLinks)-1 do
      writeln(f, relevantWords.getKeyAt(wordLinks[i].word1) + ',' + relevantWords.getKeyAt(wordLinks[i].word2) + '=' + IntToStr(wordLinks[i].strength));
   closefile(f);
end;

procedure TfrmFeeds.btnUpdateClick(Sender: TObject);
begin
   lastUserInteraction := CurrentTimeMillis;
   updateFeedDetails;
   btnUpdate.Enabled := false;
   SaveConfig;
end;

procedure TfrmFeeds.btnCheckRelClick(Sender: TObject);
var
   words: string;
   wordsList: TStringList;
begin
   timer1.Enabled := false;
   words := InputBox(self.caption, 'Enter word(s) to check separated by spaces','');
   wordsList := getCleanWordList(words);
   try
      showmessage('Simple relevance='+inttostr(checkSimpleRelevance(wordsList))+'%, ' +
         'Relationship relevance='+inttostr(checkRelevanceInRelTable(wordsList))+'%, ' +
         'Overall relevance='+inttostr(checkRelevance(words))+'%');
   finally
      wordsList.free;
      timer1.Enabled := true;
   end;
end;

procedure TfrmFeeds.enableGui(enabled: boolean);
begin
   lbFeeds.Enabled := enabled;
   btnSearchAllFeeds.Enabled := enabled;
   btnSearchSelectedFeed.Enabled := enabled;
   btnAddFeed.Enabled := enabled;
   btnDelete.Enabled := enabled;
   btnReset.Enabled := enabled;
   btnOpenFeedInBrowser.Enabled := enabled;
   btnLearn.Enabled := enabled;
   txtKeywords.Enabled := enabled;
   txtRefreshInterval.Enabled := enabled;
   btnAllActive.Enabled := enabled;
   btnAllInactive.Enabled := enabled;
   cbActive.Enabled := enabled;
   tbShowRelevance.Enabled := enabled;
   rgResultsFormat.Enabled := enabled;
   btnCancel.Enabled := not enabled;
end;


procedure TfrmFeeds.searchAll(startedManually: boolean);
var
   i: integer;
   matchingItems: Items;
begin
   timer1.Enabled := false;
   cancelled := false;
   enableGui(false);
   setLength(matchingItems,0);
   try
      log('Searching all active feeds',false);
      for i := 0 to length(feedsList)-1 do
      begin
         // ignore inactive feeds
         if not feedsList[i].active then
         begin
            log('Ignoring inactive feed ' + feedsList[i].url,false);
            continue;
         end;

         // don't scan if scanned recently
         if (currentTimeMillis - feedsList[i].lastRefresh < (feedsList[i].refreshInterval * 1000)) then
         begin
            log('Ignoring feed ' + feedsList[i].url + ' as recently scanned',true);
            continue;
         end;

         // ensure PC is idle if required, and run from timer
         if (not startedManually) and (LastWork < runAfterIdleFor) then
         begin
            Log('User is busy, waiting for idle...',false);
            waitForIdle(runAfterIdleFor);
         end;

         // get matching items for this feed and add to list of all matching items
         scanFeed(feedsList[i],matchingItems);
         Application.ProcessMessages;
         if cancelled then break;

      end;

      if not cancelled then handleFoundItems(matchingItems);

      log('Finished searching active feeds',false);
   finally
      setLength(matchingItems,0);
      enableGui(true);
      timer1.Enabled := true;
   end;
end;

{handles results from a search}
procedure TfrmFeeds.handleFoundItems(var matchingItems: Items);
begin
   if length(matchingItems) = 0 then
   begin
      log('No new matching items found',false);
      if firstTime then
      begin
         log('You may want to consider lowering the alert threshold by using the slider on the top right, it is currently set to only show articles with a relevance of at least ' + IntToStr(tbShowRelevance.position) + '%.'+chr(13)+chr(10)+
         'You could also run the learning process on more feeds - the more times you do this, the more I will learn about the type of content that interests you',false);
      end;
      exit;
   end;

   // sort & display results
   sortItems(matchingItems);

   if rgResultsFormat.ItemIndex = 0 then
      showFoundItemsOneByOne(matchingItems)
   else
      showFoundItemsAsTable(matchingItems);
end;

procedure TfrmFeeds.sortItems(var itemsToSort: Items);
var
   tmp: TItem;
   i,j: integer;
begin
   for i := 0 to length(itemsToSort) - 1 do
   begin
      for j := 0 to length(itemsToSort) - 1 do
      begin
         if itemsToSort[i].relevanceRatingFromHistory >  itemsToSort[j].relevanceRatingFromHistory then
         begin
            tmp := itemsToSort[i];
            itemsToSort[i] := itemsToSort[j];
            itemsToSort[j] := tmp;
         end;
      end;
   end;
end;

procedure TfrmFeeds.tbShowRelevanceChange(Sender: TObject);
begin
   lastUserInteraction := CurrentTimeMillis;
end;

{displays results all together in a table format}
procedure TfrmFeeds.showFoundItemsAsTable(var matchingItems: Items);
var
   frmMultiplePagesFound: TfrmMultiplePagesFound;
   i: integer;
begin
   (frmOwner as TMainForm).bringApplicationToFront;

   frmMultiplePagesFound := TfrmMultiplePagesFound.create(self);
   try
      frmMultiplePagesFound.setFeedsForm(self);
      frmMultiplePagesFound.addArticles(matchingItems);
      frmMultiplePagesFound.showModal;

      // update relevance tables
      Screen.Cursor := crHourglass;
      log('Updating relevance tables, please wait...', false);
      for i := 0 to length(matchingItems)-1 do
      begin
         frmSplashScreen.updateDetails('Updating relevance tables',((i+1) * 100) div length(matchingItems));
         updateRelevance(matchingItems[i].relevanceRatingFromUser, matchingItems[i].title + ' ' + matchingItems[i].description);

         // add URL to list of recently shown URLs
         // this means we won't show it again for a while
         addUrlToShownRecentlyList(matchingItems[i].link);

         processMessagesAtInterval;
      end;
      log('Finished updating relevance tables', false);
      Screen.Cursor := crDefault;

   finally
      frmMultiplePagesFound.free;
      frmSplashScreen.Hide;
   end;

   show; // if application was minimized, this form gets hidden which looks a bit strange
end;


{displays results one-at-a-time}
procedure TfrmFeeds.showFoundItemsOneByOne(var matchingItems: Items);
var
   i,rated: integer;
   resp: TUserResponse;
begin
   rated := 0;
   for i := 0 to length(matchingItems) -1 do
   begin
      log('Showing ' + matchingItems[i].title + ', link=' + matchingItems[i].link,true);
      resp := MainForm.alertUser(matchingItems[i].header, matchingItems[i].title, matchingItems[i].description,matchingItems[i].link, true);
      matchingItems[i].relevanceRatingFromUser := resp.relevance;

      // add URL to list of recently shown URLs
      // this means we won't show it again for a while
      addUrlToShownRecentlyList(matchingItems[i].link);

      // drop out if the user doesn't want any more alerts for a while,
      if resp.hadEnough then break;
   end;

   Screen.Cursor := crHourglass;
   // update relevance tables
   log('Updating relevance tables, please wait...', false);
   for i := 0 to length(matchingItems) -1 do
   begin
      frmSplashScreen.updateDetails('Updating relevance tables',((i+1) * 100) div length(matchingItems));
      updateRelevance(matchingItems[i].relevanceRatingFromUser, matchingItems[i].title + ' ' + matchingItems[i].description);
      if matchingItems[i].relevanceRatingFromUser <> 0 then inc(rated);
      processMessagesAtInterval;
   end;
   frmSplashScreen.Hide;
   log('Finished updating relevance tables', false);
   Screen.Cursor := crDefault;

   if (showNagScreens) and (not resp.hadEnough) and (length(matchingItems) > 1) and (rated <= length(matchingItems) div 3) then
      ShowMessage('You only rated ' + IntToStr(rated) + '/' + IntToStr(length(matchingItems)) + ' articles.' + chr(13) + chr(10) +
      'Please try to rate articles by relevance as this is the only way I can learn about what interests you and show you relevant content');

   show; // if application was minimized, this form gets hidden which looks a bit strange
end;


procedure TfrmFeeds.Timer1Timer(Sender: TObject);
begin
   // don't run if any user action in last 30 seconds
   // this stops the program starting a scan while the user is typing
   if CurrentTimeMillis - lastUserInteraction < 30000 then exit;

   try
      cbCommonIntervals.Visible := false;
      searchAll(false);

      // save config at regular intervals
      if currentTimeMillis - lastSavedConfig > saveConfigInterval then
      begin
         SaveConfig;
         SaveData;
         lastSavedConfig := CurrentTimeMillis;
      end;
   except
      on e:exception do
      begin
         if e.Message <> '' then
            log(e.Message,false)
         else
            log(e.ClassName,false);
      end;
   end;
end;

procedure TfrmFeeds.cbActiveClick(Sender: TObject);
begin
   lastUserInteraction := CurrentTimeMillis;
   if lbFeeds.ItemIndex > -1 then updateFeedDetails;
end;

procedure TfrmFeeds.cbCommonIntervalsChange(Sender: TObject);
var
   p1, p2: integer;
begin
   p1 := pos('(', cbCommonIntervals.Text);
   p2 := pos(')', cbCommonIntervals.Text);
   txtRefreshInterval.Text := copy(cbCommonIntervals.Text, p1+1, p2-p1-1);
   lastUserInteraction := CurrentTimeMillis;
   updateFeedDetails;
end;

procedure TfrmFeeds.cbCommonIntervalsExit(Sender: TObject);
begin
   cbCommonIntervals.Visible := false;
end;

{scans a feed by downloading the XML document and starting the scan}
procedure TfrmFeeds.scanFeed(var feed: TFeed; var listToAddTo:Items);
const
   RSS_START = 'ITEM';
   TITLE = 'TITLE';
   RSS_CONTENT = 'DESCRIPTION';
   ATOM_START = 'ENTRY';
   ATOM_CONTENT = 'CONTENT';
   LINK = 'LINK';
   ATOM_HREF = 'HREF';
var
   MyXml: TXmlParser;
   item: TItem;
   i: integer;
begin
   Screen.Cursor := crHourglass;
   try
      log('Scanning feed ' + feed.url,false);
      feed.lastRefresh := currentTimeMillis;
      downloadFile(feed.url);
      MyXml := TXmlParser.Create;
      try
         MyXml.LoadFromFile (downloadDir + '\' + tempFile);
         MyXml.StartScan;
         while (not cancelled) and (MyXml.Scan) do
         begin
            if uppercase(MyXml.CurName) = RSS_START then
            begin
               // RSS
               repeat
                  if not MyXml.Scan then break; // just in case
                  if uppercase(MyXml.CurName) = TITLE then
                     item.title := MyXml.CurContent
                  else if uppercase(MyXml.CurName) = RSS_CONTENT then
                     item.description := MyXml.CurContent
                  else if uppercase(MyXml.CurName) = LINK then
                     item.link := MyXml.CurContent;
               until uppercase(MyXml.CurName) = RSS_START;
               doScan(feed,item,listToAddTo);
            end
            else if uppercase(MyXml.CurName) = ATOM_START then
            begin
               // ATOM
               repeat
                  if not MyXml.Scan then break; // just in case
                  if uppercase(MyXml.CurName) = TITLE then
                     item.title := MyXml.CurContent
                  else if uppercase(MyXml.CurName) = ATOM_CONTENT then
                     item.description := MyXml.CurContent
                  else if uppercase(MyXml.CurName) = LINK then
                  begin
                     for i := 0 to MyXml.CurAttr.Count-1 do
                     begin
                        // log(MyXml.CurAttr.Name(i) + '=' + MyXml.CurAttr.Value(i));
                        if uppercase(MyXml.CurAttr.Name(i)) = ATOM_HREF then
                        begin
                           item.link := MyXml.CurAttr.Value(i);
                           break;
                        end;
                     end;
                  end;
               until uppercase(MyXml.CurName) = ATOM_START;
               doScan(feed,item,listToAddTo);
            end;
         end;
      finally
         MyXml.Free;
      end;
      log('Finished scanning feed ' + feed.url,false);
   finally
      Screen.Cursor := crDefault;
   end;
end;

{
returns a value between -100 and +100 indicating how relevant a piece of text is
-100 = totally irrelevant
0 = neutral
+100 = very relevant
}
function TfrmFeeds.checkRelevance(theText: string): integer;
var
   words: TStringList;
begin
   log('Checking relevance', true);
   result := 0;
   words := getCleanWordList(theText);
   try
      if words.Count = 0 then exit; // avoids division by zero at the end

      // first we see how popular this word is
      // i.e. how many times it has appeared in an article that
      // was rated + or - by the user
      result := checkSimpleRelevance(words);

      // only perform deep word scan if result from the basic search is positive
      // it's very unlikely we would get back a positive relationship result from words
      // that were all negative, so we can save the overhead of a deep scan
      if result > 0 then
      begin
         // now figure in the results from the relationship table
         // this checks to see whether words in this text appeared together
         // previously in articles rated + or - by the user
         // we add this to the previous result, then divide by 2 to get the final result
         result := (result + checkRelevanceInRelTable(words)) div 2;
         log('Relationship relevance=' + IntToStr(result) + '%', true);
         // result should be between -100 and +100
         assert((result <= 100) and (result >= -100));
      end;
   finally
      words.free;
   end;
end;


{performs a simple relevance check on the word weightings table}
function TfrmFeeds.checkSimpleRelevance(var words: TStringList): integer;
var
   i,j,r: integer;
begin
   result := 0;
   for i := 0 to words.count-1 do
   begin
      processMessagesAtInterval;
      j := relevantWords.IndexOf(words[i]);
      if j >= 0 then
      begin
         // get word weighting as a % of the max word weight
         r := (relevantWords.getValueAt(j)*100) div maxWordWeight;
         result := result + r;
      end;
   end;
   result := result div words.count;
   log('Relevance=' + IntToStr(result) + '%', true);

   // result should be between -100 and +100
   assert((result <= 100) and (result >= -100));
end;


{
returns a value between -100 and +100 indicating how relevant a piece of text is
using the relationship table, i.e. how well the words relate to each other
-100 = totally unrelated, i.e. words have not appeared together in text rated highly by the user,
or appeared together in text rated negatively by the user
0 = neutral
+100 = very related, i.e. words have appeared together in text rated highly by the user
}
function TfrmFeeds.checkRelevanceInRelTable(var words: TStringList): integer;
var
   wordsFound,i,j,k,index1,index2,r: integer;
   start: COMP;
   cacheArray: array of integer;
begin
   result := 0;
   wordsFound := 0;
   removeDuplicates(words);
   if words.Count < 2 then exit;
   log('Checking ' + IntToStr(words.count) + ' words against word relationship table', true);
   start := currentTimeMillis;
   result := 0;

   // to save continually looking up the word indices, look them up
   // once & save in an array
   setLength(cacheArray,words.count);
   for i := 0 to words.count-1 do
      cacheArray[i] := relevantWords.indexOf(words[i]);

   for i := 0 to words.count-1 do
   begin
      processMessagesAtInterval;
      index1 := cacheArray[i];
      if index1 = -1 then continue;
      inc(wordsFound);
      for j:= i to words.count-1 do
      begin
         if i = j then continue;
         index2 := cacheArray[j];
         if index2 = -1 then continue;

         // get relationship weight between these words as a % of the max word weight
         k := findLinkIndex(index1, index2);
         if (k <> -1) then
            r := (wordLinks[k].strength*100) div maxWordWeight
         else
            r := 0;

         // log('Rel between ' + words[i] + ' & ' + words[j] + ' = ' + IntToStr(r),false);

         // we multiply by 2 here because we are only scanning half the words for efficiency
         // i.e. if we know that the rel between IRAN and NUCLEAR is 5, there is no need to
         // check the rel between NUCLEAR and IRAN because it will be the same
         // however, we still add it on so that the result would be identical
         // to result from doing a full scan of all elements in the table
         result := result + (r*2);
      end;
   end;

   log('Word relationship scan of ' + IntToStr(words.count) + ' words took ' + FloatToStr(currentTimeMillis-start), true);

   if wordsFound < 2 then exit; // prevent division by zero

   // divide total by number of possible combinations
   // result := result div (words.count*(words.count-1));

   // CHANGE - we now only count the words we know about
   // technically this gives a less "pure" result, but otherwise
   // the result is always 0% or 1% which is useless
   result := result div (wordsFound*(wordsFound-1));

   // final result should be between -100 and +100
   assert((result <= 100) and (result >= -100));
end;


{scans a feed for items from an article
we scan for keywords, then for relevance}
procedure TfrmFeeds.doScan(var feed: TFeed; var item: TItem; var itemList: Items);
var
   matchingLine: string;
   relevance: integer;
begin
   // ignore any links that have already been shown
   if urlShownRecently(item.link) then
   begin
      log('URL ' + item.link + ' has been shown recently, ignoring it',true);
      exit;
   end;

   // clean up first
   item.title := removeHtml(item.title);
   item.description := removeHtml(item.description);

   // ignore any identical items already in the list
   if itemInList(item, itemList) then
   begin
      log('Duplicate URL ' + item.link + ' found, ignoring it',true);
      exit;
   end;

   // do keyword search first (title and description)
   matchingLine := scanKeywords(feed,item.title + ' ' + item.description);

   // if no keyword matches, search on overall relevance
   if matchingLine <> '' then
   begin
      if learning then
      begin
         item.header := 'Learning';
      end
      else
      begin
         item.header := 'Keyword match on ' + matchingLine;
         log(item.header, false);
      end;
      item.relevanceRatingFromHistory := 100;
      setLength(itemList, length(itemList)+1);
      itemList[length(itemList)-1] := item;
   end
   else
   begin
      relevance := checkRelevance(item.title + ' ' + item.description);
      if relevance >= tbShowRelevance.position then
      begin
         item.header := 'Article found, relevance=' + IntToStr(relevance) + '%';
         log(item.header, false);
         item.relevanceRatingFromHistory := relevance;
         setLength(itemList, length(itemList)+1);
         itemList[length(itemList)-1] := item;
      end;
   end;
end;

{
scans a particular feed for keywords
if keywords found, the function returns the line that matched
if no key ords found, the function returns ''
}
function TfrmFeeds.scanKeywords(var feed: TFeed; textToSearch: string): string;
var
   keywordsOnALine: TStringList;
   i,j,found: integer;
   tmpKeywords: string;
begin
   {keywords is a list of strings in the format:
   '1:gary,mckinnon,pentagon'
   '1:jordan,arabs,middle east,!katie price

   all keywords must be found on a line for a match to be found
   }

   result := '';

   for i := 0 to feed.keywords.count-1 do
   begin
      // ignore keyword lines that don't start with '1:'
      if copy(feed.keywords[i], 1, 2) <> '1:' then
      begin
         log('Ignoring inactive keywords: ' + feed.keywords[i],true);
         continue;
      end;

      // remove '1:'
      tmpKeywords := uppercase(copy(feed.keywords[i], 3, length(feed.keywords[i])-2));

      // ignore blank lines
      if length(tmpKeywords) = 0 then
      begin
         log('No keywords',true);
         continue;
      end;

      log('Searching for ' + tmpKeywords,true);

      // break up each line into keywords
      keywordsOnALine := parseString(tmpKeywords, ',');
      try
         // see if each keyword was found on this line
         found := 0;
         for j := 0 to keywordsOnALine.Count-1 do
         begin
            // see if word must *not* be found, e.g. !iran
            if copy(keywordsOnALine[j], 1, 1) = '!' then
            begin
               if pos(copy(keywordsOnALine[j],2, length(keywordsOnALine[j])-1),uppercase(textToSearch)) > 0 then
               begin
                  found := -1;
                  break;
               end
               else
                  inc(found);
            end
            else
            begin
               // see if word *must* be found
               if pos(keywordsOnALine[j],uppercase(textToSearch)) > 0 then inc(found);
            end;
         end;
         if found = keywordsOnALine.count then
         begin
            result := tmpKeywords;
            exit;
         end
         else
            log('No match found for ' + tmpKeywords,true);
      finally
         keywordsOnALine.free;
      end;
   end;

end;


procedure TfrmFeeds.txtAddFeedURLChange(Sender: TObject);
begin
   lastUserInteraction := CurrentTimeMillis;
   btnUpdate.Enabled := false;
   if not updatingGUI then
   begin
      lbFeeds.itemIndex := -1; // deselect so we don't overwrite the selected item
      cbActive.Checked := true;
      txtKeywords.Text := '';
      txtRefreshInterval.Text := IntToStr(defaultRefreshInterval);
   end;
end;

procedure TfrmFeeds.txtAddFeedURLClick(Sender: TObject);
begin
   lastUserInteraction := CurrentTimeMillis;
end;

procedure TfrmFeeds.txtKeywordsChange(Sender: TObject);
begin
   if lbFeeds.ItemIndex > -1 then btnUpdate.Enabled := true;
   lastUserInteraction := CurrentTimeMillis;
end;

procedure TfrmFeeds.txtKeywordsClick(Sender: TObject);
begin
   lastUserInteraction := CurrentTimeMillis;
end;

procedure TfrmFeeds.txtKeywordsExit(Sender: TObject);
begin
   validateKeywords(txtKeywords);
   updateFeedDetails;
end;

procedure TfrmFeeds.txtRefreshIntervalChange(Sender: TObject);
begin
   lastUserInteraction := CurrentTimeMillis;
   if lbFeeds.ItemIndex > -1 then btnUpdate.Enabled := true;
end;


procedure TfrmFeeds.txtRefreshIntervalEnter(Sender: TObject);
var
   i: integer;
begin
   cbCommonIntervals.Visible := true;
   for i := 0 to cbCommonIntervals.Items.Count-1 do
   begin
      if pos('('+ txtRefreshInterval.text+')', cbCommonIntervals.Items[i]) > 0 then
      begin
         cbCommonIntervals.ItemIndex := i;
         exit;
      end;
   end;
end;

procedure TfrmFeeds.txtRefreshIntervalExit(Sender: TObject);
begin
   updateFeedDetails;
end;

procedure TfrmFeeds.updateFeedDetails;
begin
   if (not updatingGUI) and (lbFeeds.itemIndex > -1) then
   begin
      lastUserInteraction := CurrentTimeMillis;
      with feedsList[lbFeeds.itemIndex] do
      begin
         active := cbActive.Checked;
         keywords.free; // free previous list
         keywords := parseString(txtKeywords.Text,chr(13)+chr(10));
         try
            refreshInterval := StrToInt(txtRefreshInterval.text);
            if refreshInterval < MIN_REFRESH_VALUE then raise Exception.create('Invalid refresh value');
         except
            Log('Invalid refresh interval, setting to default value ' + IntToStr(MIN_REFRESH_VALUE),false);
            refreshInterval := MIN_REFRESH_VALUE;
         end;
      end;
      log('Feed details updated',false);
   end;
end;



{return result is an empty string if successful, or an error message if unsuccessful}
function TfrmFeeds.downloadFile(url: string): string;
var
   HTTP: TIdHTTP;
   Stream: TMemoryStream;
   localName: string;
begin
   log('Downloading ' + url,true);
   result := '';
   try
      url := urlencode(url);
      localName := downloadDir + '\' + tempFile;
      DeleteFile(localName);
      HTTP := TIdHTTP.Create(nil);
      try
         with HTTP do
         begin
            with ProxyParams do
            begin
               ProxyServer := (frmOwner as TMainForm).proxyDetails.ProxyServer;
               ProxyPort := (frmOwner as TMainForm).proxyDetails.ProxyPort;
               ProxyUsername := (frmOwner as TMainForm).proxyDetails.ProxyUsername;
               ProxyPassword := (frmOwner as TMainForm).proxyDetails.ProxyPassword;
            end;
            HandleRedirects := True;
            AllowCookies := false;
            ReadTimeout := SOCKET_TIMEOUT;
            ConnectTimeout := SOCKET_TIMEOUT;
            Request.UserAgent := 'Mozilla/4.0';
            Request.Connection := 'Keep-Alive';
            Request.ProxyConnection := 'Keep-Alive';
            Request.CacheControl := 'no-cache';
         end;
         Stream := TMemoryStream.Create;
         try
            HTTP.Get(Url,Stream);
            Stream.SaveToFile(localName);
            log('Downloaded ' + url + ' ok',true);
         finally
            Stream.Free;
         end;
      finally
         HTTP.Free;
      end;

   except
      on e:exception do
      begin
         if e.Message <> '' then
            result := e.Message
         else
            result := e.ClassName;

         log('Error downloading ' + url + ': ' + result,false);
      end;
   end;
end;

procedure TfrmFeeds.FormActivate(Sender: TObject);
begin
   if (FirstTime) and (not shownHelp) then
   begin
      Timer1.Enabled := false;
      shownHelp := true;
      ShowMessage('To start you off I have added some example RSS feeds.  You can add/remove feeds using the controls at the top of this page.' + chr(13)+chr(10) + chr(13)+chr(10) +
      'When you have set up your feeds, use the control on the right of the page to set the trigger keywords.'+ chr(13)+chr(10) +
      'Note that keywords should be comma-separated, e.g. ''1:iran,nuclear,inspection'' - all keywords that are on a separate line must be found for the alert to be shown, but you can define as many lines as you like, e.g.' + chr(13)+chr(10) +chr(13)+chr(10) +
      '1:iran,nuclear,inspection'+chr(13)+chr(10) +
      '1:economy,inflation,interest rates'+chr(13)+chr(10) +
      '1:cricket tour,england'+chr(13)+chr(10)+
      '1:film,!cartoon,!animation'+chr(13)+chr(10) +chr(13)+chr(10) +
      'Note that keywords can contain spaces, and you can use the ''!'' character to specify that a word must NOT be found. The ''1:'' at the beginning of each line simply defines this keyword line as ''active'''+chr(13)+chr(10) +chr(13)+chr(10) +
      'If you would like me to learn about your interests so that I can find relevant content without you having to set up keywords , please select a feed and click Learn, then rate each article''s relevance from ''Totally irrelvant'' to ''Very relevant''' + chr(13)+chr(10) +
      'Articles will be shown one at a time, but if you prefer to see them all together you can change the ''Results Format'' to ''As a table'''+chr(13)+chr(10) +chr(13)+chr(10) +
      'Feeds marked as ''active'' are scanned at defined intervals, but you can also scan a feed manually using the ''Search'' buttons at the top');

      Timer1.Enabled := true;
   end;

end;

procedure TfrmFeeds.FormClose(Sender: TObject; var Action: TCloseAction);
begin
   feedsFormShowing := false;
end;

procedure TfrmFeeds.FormCreate(Sender: TObject);
begin
   cancelled := false;
   cbCommonIntervals.DropDownCount := cbCommonIntervals.Items.Count;
   lastUserInteraction := CurrentTimeMillis;
   lblShowRelevance.Hint := 'As you provide feedback on each article''s relevance, Webu will learn about the kind of articles you find interesting.' + chr(13) + chr(10) +
      'You can use this slider to indicate how relevant an article needs to be before Webu decides to show it to you.' + chr(13) + chr(10) +
      'If you move the slider towards LOW, you will get lots of articles that may have little relevance.' + chr(13) + chr(10) +
      'However, if you move the slider towards HIGH, Webu will only show you articles it thinks are very relevant.' + chr(13) + chr(10) +
      'Do not set it to either extreme though or you will get every single article or none at all.' + chr(13) + chr(10) +
      'Try starting off by setting it somewhere in the middle and adjusting it slowly.';
   tbShowRelevance.Hint := lblShowRelevance.Hint;
   label3.Hint := lblShowRelevance.Hint;
   Timer1.Interval := MIN_REFRESH_VALUE * 1000;
   // Timer1.Interval := 15000; // for testing
   txtRefreshInterval.Text := IntToStr(MIN_REFRESH_VALUE);
   lblRefreshInterval.Caption := lblRefreshInterval.Caption + IntToStr(MIN_REFRESH_VALUE) + ')';
   downloadDir := ExtractFilePath(Application.ExeName) + 'temp';
   dataDir := ExtractFilePath(Application.ExeName) + 'data';
   forceDirectories(downloadDir);
   forceDirectories(dataDir);
   forceDirectories(dataDir + '\backup');
   urlsShownRecently := TCompMap.create;
   relevantWords := TShortIntMap.create;
   commonWords := TStringList.create;
   commonWords.Sorted := true;
   commonWords.CaseSensitive := false;
end;


procedure TfrmFeeds.FormDestroy(Sender: TObject);
begin
   urlsShownRecently.free;
   relevantWords.free;
   commonWords.Free;
end;

procedure TfrmFeeds.FormHide(Sender: TObject);
begin
   feedsFormShowing := false;
end;

procedure TfrmFeeds.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
   // note - form's "KeyPreview" must be set to true or this event isn't fired
   btnUnitTests.visible := ((ssShift in Shift) and (ssAlt in Shift));
   btnCheckRel.visible := ((ssShift in Shift) and (ssAlt in Shift));
end;

procedure TfrmFeeds.lbFeedsClick(Sender: TObject);
var
   feed: TFeed;
begin
   lastUserInteraction := CurrentTimeMillis;
   cbCommonIntervals.Visible := false;
   updatingGUI := true; // so we don't fire the OnChange events
   try
      feed := feedsList[lbFeeds.itemIndex];
      txtAddFeedURL.Text := feed.url;
      txtRefreshInterval.text := IntToStr(feed.refreshInterval);
      txtKeywords.Text := unparseString(feed.keywords, chr(13)+chr(10));
      cbActive.Checked := feed.active;
      btnUpdate.Enabled := false;
   finally
      updatingGUI := false;
   end;
end;

procedure TfrmFeeds.loadConfig(confFile: TIniFile);
var
   feedCount,i: integer;
   storedKeywords: string;
begin
   frmSplashScreen.updateDetails('Loading config...',0);

   // count the feeds first
   feedCount := confFile.readInteger('main','feedCount',0);

   // create array & read feeds into it
   setLength(feedsList, feedCount);
   for i := 0 to feedCount-1 do
   begin
      frmSplashScreen.updateDetails('Loading feeds...',((i+1) * 100) div feedCount);
      feedsList[i].url := confFile.ReadString('rss','feed.' + IntToStr(i) + '.url','');
      feedsList[i].active := confFile.ReadBool('rss','feed.' + IntToStr(i) + '.active',true);
      storedKeywords := StringReplace(confFile.ReadString('rss','feed.' + IntToStr(i) + '.keywords',''), 'CRLF', chr(13)+chr(10), [rfReplaceAll, rfIgnoreCase]);
      feedsList[i].keywords := parseString(storedKeywords,chr(13)+chr(10));
      feedsList[i].refreshInterval := confFile.ReadInteger('rss','feed.' + IntToStr(i) + '.refreshInterval',defaultRefreshInterval);
      feedsList[i].lastRefresh := StrToFloat(confFile.ReadString('rss','feed.' + IntToStr(i) + '.lastRefresh','0'));
      lbFeeds.Items.Add(feedsList[i].url);
   end;

   if length(feedsList) > 0 then
   begin
      lbFeeds.ItemIndex := 0;
      lbFeeds.OnClick(self);
   end;

   lastSavedConfig := CurrentTimeMillis;

    // other props
   minWordLength := confFile.ReadInteger('main','minWordLength',4);
   maxWordLength := confFile.ReadInteger('main','maxWordLength',12);
   maxWordWeight := confFile.ReadInteger('main','maxWordWeight',300);
   assert(maxWordWeight <= high(relType));
   saveConfigInterval := confFile.ReadInteger('main','saveConfigInterval',10) * MINUTE;

   // load the list of urls we have shown to the user
   urlsShownRecently.loadFromFile(dataDir + '\' + shownUrlsFile);

   // common words - these are ignored when doing relevance scans
   commonWords.LoadFromFile(dataDir + '\commonwords.dat');

   // other props
   tbShowRelevance.position := confFile.ReadInteger('main','showRelevance',30);
   rgResultsFormat.ItemIndex := confFile.ReadInteger('main','resultsFormat',0);

   // load list of word weightings
   frmSplashScreen.updateDetails('Loading word weightings...',0);
   loadWordWeightingsTable(dataDir + '\' + wordWeightingsFile);

   // load word relationship table
   setLength(wordLinks, 0);
   if relevantWords.count > 0 then
   begin
      // check for old relationship table (from previous versions)
      if FileExists(dataDir + '\' + oldRelTableFile) then
      begin
         loadOldRelTable(dataDir + '\' + oldRelTableFile);
      end
      else
      begin
         if FileExists(dataDir + '\' + linksTableFile) then
            loadLinksTable;
      end;
   end;

   // remove any words not in clean words list
   // as this list may have been improved since data was saved
   /// DON'T REALLY NEED THIS NOW
   // cleanData;

   frmSplashScreen.Hide;
   Timer1.enabled := true;
end;


{remove any words/links not in clean words list}
procedure TfrmFeeds.cleanData;
var
   i,j,k: integer;
   allWords: string;
   cleanWords: TStringList;
begin
   allWords := '';
   for i := 0 to relevantWords.count-1 do
      allWords := allWords + relevantWords.getKeyAt(i) + ' ';

   cleanWords := getCleanWordList(allWords);
   try
      for i := relevantWords.count-1 downto 0 do
      begin
         frmSplashScreen.updateDetails('Checking data...',relevantWords.count - i + 1,relevantWords.count,10);
         if cleanWords.IndexOf(relevantWords.getKeyAt(i)) = -1 then
         begin
            // showmessage('removing ' + relevantWords.getKeyAt(i) + ', index ' + inttostr(i));
            relevantWords.delete(i);
            for j := length(wordLinks)-1 downto 0 do
            begin
               if (wordLinks[j].word1 = i) or (wordLinks[j].word2 = i) then
               begin
                  for k := j to length(wordLinks)-2 do
                     wordLinks[k] := wordLinks[k+1];
                  setLength(wordLinks, length(wordLinks)-1);
               end;
            end;
         end;
      end;
   finally
      cleanWords.free;
      frmSplashScreen.Hide;
   end;
end;

{loads the word weightings table from a file}
procedure TfrmFeeds.loadWordWeightingsTable(filename: string);
var
   i,r: integer;
begin
   relevantWords.loadFromFile(filename);

   // ensure all words are within min/max range
   for i := 0 to relevantWords.count-1 do
   begin
      r := relevantWords.getValueAt(i);
      if r > maxWordWeight then
         relevantWords.setValueAt(i,maxWordWeight)
      else if r < 0-maxWordWeight then
         relevantWords.setValueAt(i,0-maxWordWeight);
   end;
end;


{loads the word relationship table from a file}
function TfrmFeeds.loadOldRelTable(filename: string): boolean;
var
   i,j: integer;
   tmpRelTable: T2DRelTypeArray;
begin
   result := false;
   if not FileExists(filename) then exit;

   try
      Load2DShortArray(tmpRelTable, filename);

      // check length is the same as the relevant words table
      if (length(tmpRelTable) = relevantWords.count) and
         (length(tmpRelTable[relevantWords.count-1]) = relevantWords.count) then

         result := true;

      // create new array of links from the old reltable
      for i := 0 to length(tmpRelTable) - 1 do
      begin
         for j:= 0 to length(tmpRelTable) - 1 do
         begin
            if i = j then continue;
            if tmpRelTable[i,j] = 0 then continue;
            setLength(wordLinks, length(wordLinks)+1);
            with wordLinks[length(wordLinks)-1] do
            begin
               word1 := i;
               word2 := j;
               strength := tmpRelTable[i,j];
            end;
         end;
      end;

      dumpLinkTable;
         
   except
      on e:exception do
      begin
         log(e.ClassName + ': ' + e.Message,false);
      end;
   end;
end;

{loads the word link table from a file}
procedure TfrmFeeds.loadLinksTable;
var
   f: textfile;
   lineCount,i,p1,p2: integer;
   line: string;
begin
   AssignFile(f, dataDir + '\' + linksTableFile);
   reset(f);

   // first pass over the file to count the lines
   lineCount := 0;
   while not eof(f) do
   begin
      readln(f, line);
      inc(lineCount);
   end;

   // now read the link data into the array
   setLength(wordLinks, lineCount);
   reset(f);
   for i := 0 to lineCount-1 do
   begin
      readln(f, line);
      p1 := pos(',',line);
      with wordLinks[i] do
      begin
         word1 := StrToInt(copy(line, 1, p1-1));
         p2 := pos('=',line);
         word2 := StrToInt(copy(line, p1+1, p2-p1-1));
         strength := StrToInt(copy(line, p2+1, length(line)-p1-1));
      end;
      frmSplashScreen.updateDetails('Loading ' + linksTableFile,i,lineCount,10);
   end;

   closefile(f);
   frmSplashScreen.hide;
end;


procedure TfrmFeeds.saveConfig(confFile: TIniFile);
var
   i: integer;
   storedKeywords: string;
begin
   frmSplashScreen.updateDetails('Saving config...',0);
   log('Saving config...',false);
   Screen.Cursor := crHourglass;

   for i := 0 to length(feedsList)-1 do
   begin
      frmSplashScreen.updateDetails('Saving feeds...',((i+1) * 100) div length(feedsList));
      confFile.WriteString(iniPrefix,'feed.' + IntToStr(i) + '.url', feedsList[i].url);
      confFile.WriteBool(iniPrefix,'feed.' + IntToStr(i) + '.active', feedsList[i].active);
      storedKeywords := unparseString(feedsList[i].keywords,chr(13)+chr(10));
      confFile.WriteString(iniPrefix,'feed.' + IntToStr(i) + '.keywords', StringReplace(storedKeywords, chr(13)+chr(10), 'CRLF', [rfReplaceAll, rfIgnoreCase]));
      confFile.WriteInteger(iniPrefix,'feed.' + IntToStr(i) + '.refreshInterval', feedsList[i].refreshInterval);
      confFile.WriteString(iniPrefix,'feed.' + IntToStr(i) + '.lastRefresh', FloatToSTr(feedsList[i].lastRefresh));
   end;

   confFile.WriteInteger('main','feedCount',length(feedsList));

   // other props
   confFile.WriteInteger('main','showRelevance',tbShowRelevance.position);
   confFile.WriteInteger('main','resultsFormat',rgResultsFormat.ItemIndex);

   log('Finished saving config...',false);
   frmSplashScreen.Hide;
   Screen.Cursor := crDefault;
end;


{saves the data i.e. word weightings, relationships etc}
procedure TfrmFeeds.saveData;
begin
   frmSplashScreen.updateDetails('Saving data...',0);
   log('Saving data...',false);
   Screen.Cursor := crHourglass;

   // save the list of urls we have shown to the user
   if backupDataFiles then
   begin
      deleteFile(dataDir + '\backup\' + shownUrlsFile);
      moveFile(pchar(dataDir + '\' + shownUrlsFile),pchar(dataDir + '\backup\' + shownUrlsFile));
   end;
   urlsShownRecently.saveToFile(dataDir + '\' + shownUrlsFile);

   // save the word weightings file
   if backupDataFiles then
   begin
      deleteFile(dataDir + '\backup\' + wordWeightingsFile);
      moveFile(pchar(dataDir + '\' + wordWeightingsFile),pchar(dataDir + '\backup\' + wordWeightingsFile));
   end;
   relevantWords.saveToFile(dataDir + '\' + wordWeightingsFile);

   // save the word links file
   if backupDataFiles then
   begin
      deleteFile(dataDir + '\backup\' + linksTableFile);
      moveFile(pchar(dataDir + '\' + linksTableFile),pchar(dataDir + '\backup\' + linksTableFile));
   end;
   saveLinksTable;

   // delete old relationship table, if it exists
   if FileExists(dataDir + '\' + oldRelTableFile) then DeleteFile(dataDir + '\' + oldRelTableFile);

   log('Finished saving data...',false);
   frmSplashScreen.hide;
   Screen.Cursor := crDefault;
end;


procedure TfrmFeeds.log(line: string; isVerbose: boolean);
begin
   if (not verboseOutput) and (isVerbose) then exit;
   if txtResults.Lines.Count > 10000 then txtResults.Clear;
   txtResults.Lines.Add(DateTimeToStr(Now) + ': ' + line);
   SendMessage(txtResults.handle, EM_SCROLLCARET,0,0);
end;


procedure TfrmFeeds.rgResultsFormatClick(Sender: TObject);
begin
   lastUserInteraction := CurrentTimeMillis;
end;

procedure TfrmFeeds.addUrlToShownRecentlyList(url: string);
begin
   urlsShownRecently.put(url,CurrentTimeMillis);
end;


{returns true if an item is already present in a list of items}
function TfrmFeeds.itemInList(var item:TItem; var itemList: Items): boolean;
var
   i: integer;
begin
   result := false;
   for i := 0 to length(itemList)-1 do
   begin
      if (itemList[i].title = item.title) and (itemList[i].link = item.link) then
      begin
         result := true;
         break;
      end;
   end;
end;


{returns true if a URL was shown recently}
function TfrmFeeds.urlShownRecently(url: string): boolean;
var
   i: integer;
begin
   result := false;

   for i := urlsShownRecently.Count-1 downto 0 do
   begin
      // if url is old, remove it
      if CurrentTimeMillis - urlsShownRecently.getValueAt(i) > repeatInterval then
      begin
         urlsShownRecently.Delete(i);
      end
      else
      begin
         if url = urlsShownRecently.getKeyAt(i) then
         begin
            result := true;
            exit;
         end;
      end;
   end;
end;

{updates the word-weightings using a piece of text}
procedure TfrmFeeds.updateRelevance(relevance: integer; theText: string);
var
   words: TStringList;
   i,j,r: integer;
begin
   if relevance = 0 then exit;

   log('Updating relevance', true);
   words := getCleanWordList(theText);
   try
      // go through list of words, adding/updating relevance waitings
      // each time a word is found, we increase/decrease its weighting
      // by the relevance factor specified by the user
      for i := 0 to words.count-1 do
      begin
         processMessagesAtInterval;
         j := relevantWords.IndexOf(words[i]);
         if j = -1 then
         begin
            // new word - add it to the list and set initial relevance
            relevantWords.put(words[i],relevance);
         end
         else
         begin
            // we already have this word so update its relevance
            // by adding the value indicated by the user
            // weightings have a min & max value to stop runaway values
            r := relevantWords.getValueAt(j);
            r := r + relevance;
            if r > maxWordWeight then r := maxWordWeight;
            if r < 0-maxWordWeight then r := 0-maxWordWeight;
            relevantWords.setValueAt(j,r);
         end;
      end;

      // update the word-relationship table
      updateRelTable(words, relevance);
   finally
      words.free;
   end;
end;


{returns the index in the links array of a link between 2 words
or -1 if no link exists
FIXME ought to make this more efficient, at present it just trawls the entire list}
function TfrmFeeds.findLinkIndex(index1, index2: integer): integer;
var
   i: integer;
begin
   result := -1;
   for i := 0 to length(wordLinks) - 1 do
   begin
      if (wordLinks[i].word1 = index1) and (wordLinks[i].word2 = index2) then
      begin
         result := i;
         break;
      end;
   end;
end;

{
updates the table holding the relationship between words
this works as follows: imagine the user rates the following 2 articles
THE KING OF JORDAN IS CALLED ABDULLAH (rated +5 by the user)
JORDAN ALSO KNOWN AS KATIE PRICE IS A CELEBRITY (rated -3 by the user)
the relevance table will record the fact that the user prefers articles
about King Abdullah of Jordan over articles about Jordan the celebrity
}
procedure TfrmFeeds.updateRelTable(var words: TStringList; relevance: integer);
var
   i,j,k,index1,index2: integer;
   cacheArray: array of integer;
begin
   log('Updating rel table', true);
   removeDuplicates(words);

   // to save continually looking up the word indices, look them up
   // once & save in an array
   setLength(cacheArray,words.count);
   for i := 0 to words.count-1 do
      cacheArray[i] := relevantWords.indexOf(words[i]);

   for i := 0 to words.count-1 do
   begin
      processMessagesAtInterval;
      index1 := cacheArray[i];
      for j := 0 to words.count-1 do
      begin
         if i = j then continue;
         index2 := cacheArray[j];
         k := findLinkIndex(index1, index2);
         if k > -1 then
         begin
            // link already exists, so update the strength
            with wordLinks[k] do
            begin
               strength := strength + relevance;
               if strength > maxWordWeight then strength := maxWordWeight;
               if strength < 0-maxWordWeight then strength := 0-maxWordWeight;
            end;
         end
         else
         begin
            // add new link
            setLength(wordLinks, length(wordLinks)+1);
            with wordLinks[length(wordLinks)-1] do
            begin
               word1 := index1;
               word2 := index2;
               strength := relevance;
            end;
         end;
      end;
   end;
end;


{cleans up a piece of text by removing HTML, punctuation, common words, small words etc.}
function TfrmFeeds.getCleanWordList(theText: string): TStringList;
var
   i: integer;
begin
   theText := trim(uppercase(removeHTML(theText)));

   // replace some common grammar
   theText := StringReplace(theText, chr(39) + 'S', '', [rfReplaceAll, rfIgnoreCase]);

   // remove all other apostrophes
   theText := StringReplace(theText, '''', '', [rfReplaceAll, rfIgnoreCase]);

   // rip out any non-alphabetic characters/spaces
   theText := removeNonAlphabetic(theText);

   // split up into words
   result := parseString(theText, ' ');

   // clean up the list of words
   for i := result.count-1 downto 0 do
   begin
      result[i] := trim(result[i]);

      // remove words that are too short or too long
      if (length(result[i]) < minWordLength) or (length(result[i]) > maxWordLength) then
      begin
         result.Delete(i);
         continue;
      end;

      // because we don't want separate weightings for APPLE and APPLES,
      // we remove any trailing 'S' from the word (if it's long enough)
      // unfortunately this may cause things like "JONES" to become "JONE"
      // but this shouldn't affect the ratings, unless there are words that
      // have a totally different meaning when the 'S' is removed
      // (I can't think of any, but if you can then please let me know)
      // (exclude words ending in double S, e.g. CROSS, MESS
      if (length(result[i]) > minWordLength) and (copy(result[i], length(result[i]), 1) = 'S')  and (copy(result[i], length(result[i])-1, 1) <> 'S') then
         result[i] := copy(result[i], 1, length(result[i])-1);

      // remove any common words such as AND, THE, etc
      if (findStringInSortedList(result[i],commonwords)) then
         result.Delete(i);
   end;
end;


function TfrmFeeds.getRelevantWordCount: integer;
begin
   result := relevantWords.count;
end;

end.

