unit pagemonitor;

interface

// functions found on
// http://www.cryer.co.uk/brian/delphi/twebbrowser/read_write_form_elements.htm#SetValueOfField


uses  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, OleCtrls, SHDocVw, StdCtrls, mshtml, ExtCtrls, IniFiles,showbigtext,utils,pagefound,
  stringmap;


type
  TfrmPageMonitor = class(TForm)
    Panel1: TPanel;
    Panel3: TPanel;
    WebBrowser1: TWebBrowser;
    Timer1: TTimer;
    Panel2: TPanel;
    Label4: TLabel;
    txtResults: TMemo;
    Panel4: TPanel;
    Label1: TLabel;
    txtAddress: TEdit;
    btnGo: TButton;
    btnShowRawText: TButton;
    lblRefreshInterval: TLabel;
    txtRefreshInterval: TEdit;
    Splitter1: TSplitter;
    Panel5: TPanel;
    Label3: TLabel;
    txtAlerts: TMemo;
    txtStartTime: TEdit;
    txtEndTime: TEdit;
    Label2: TLabel;
    Label5: TLabel;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure btnGoClick(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure txtRefreshIntervalChange(Sender: TObject);
    procedure WebBrowser1DocumentComplete(ASender: TObject;
      const pDisp: IDispatch; var URL: OleVariant);
    procedure btnShowRawTextClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure WebBrowser1NavigateError(ASender: TObject; const pDisp: IDispatch;
      var URL, Frame, StatusCode: OleVariant; var Cancel: WordBool);
    procedure txtAlertsExit(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
    frmOwner: TForm;
    linesShownRecently: TStringMap;
    function getPageContents: string;
    procedure startSearch;
    function search(searchLines: TStrings; textToSearch: string): TStringList;
    procedure addLineToShownRecentlyList(line: string);
    function lineShownRecently(line: string): boolean;
    procedure log(line: string; isVerbose: boolean);
    function isTimeToScan:boolean;
  public
    { Public declarations }
    procedure setOwner(_frmOwner: TForm);
    procedure getWebPage;
    procedure loadConfig(prefix: string; var IniFile: TIniFile);
    procedure saveConfig(prefix: string; var IniFile: TIniFile);
  end;

implementation

uses Main;

{$R *.dfm}

procedure TfrmPageMonitor.btnShowRawTextClick(Sender: TObject);
var
  contents: string;
begin
   contents := getPageContents;
   if length(contents) > 0 then
      showBigTextForm(contents)
   else
      ShowMessage('Nothing to display');
end;

procedure TfrmPageMonitor.btnGoClick(Sender: TObject);
var
   currentTimerState: boolean;
begin
   currentTimerState := Timer1.enabled; // suspend timer temporarily
   Timer1.enabled := false;
   try
      getWebPage;
   finally
      Timer1.enabled := currentTimerState; // reset timer
   end;
end;

procedure TfrmPageMonitor.FormClose(Sender: TObject; var Action: TCloseAction);
begin
   Action := caFree;
end;

procedure TfrmPageMonitor.FormCreate(Sender: TObject);
begin
   linesShownRecently := TStringMap.create;
   txtRefreshInterval.Text := IntToStr(MIN_REFRESH_VALUE);
   lblRefreshInterval.Caption := lblRefreshInterval.Caption + IntToStr(MIN_REFRESH_VALUE) + ')';
end;

procedure TfrmPageMonitor.FormDestroy(Sender: TObject);
begin
   linesShownRecently.free;
end;




function TfrmPageMonitor.isTimeToScan:boolean;
var
   start,finish: TDate;
   yy,mm,dd: word;
begin
   decodeDate(Now, yy, mm, dd);
   start := encodeDate(yy,mm,dd) + getTimeFromString(txtStartTime.text);
   finish := encodeDate(yy,mm,dd) + getTimeFromString(txtEndTime.text);
   result := (Now >= start) and (Now <= finish);
end;

procedure TfrmPageMonitor.Timer1Timer(Sender: TObject);
begin
   if isTimeToScan then
   begin
      if LastWork < runAfterIdleFor then
      begin
         Log('User is busy, waiting for idle...',false);
         waitForIdle(runAfterIdleFor);
      end;
      processMessagesAtInterval;
      getWebPage;
   end
   else
      log('Not time to scan',true);
end;


procedure TfrmPageMonitor.getWebPage;
begin
   if length(txtAddress.text) = 0 then exit;

   log('Downloading ' + txtAddress.text, false);
   try
      btnGo.enabled := false;
      try
         WebBrowser1.Navigate(txtAddress.text);
      finally
         btnGo.enabled := true;
      end;
   except
      on e:exception do
      begin
         if e.Message <> '' then
            log(e.Message,false)
         else
            log(e.ClassName,false);
      end;
   end;
end;

procedure TfrmPageMonitor.txtAlertsExit(Sender: TObject);
begin
   validateKeywords(txtAlerts);
end;

procedure TfrmPageMonitor.txtRefreshIntervalChange(Sender: TObject);
var
   interval: integer;
begin
   try
      interval := StrToInt(txtRefreshInterval.text);
   except
      interval := 0;
   end;

   if interval >= MIN_REFRESH_VALUE then
   begin
      timer1.Interval := interval * 1000;
      timer1.Enabled := true;
      log('Enabled auto-check',false);
   end
   else
   begin
      timer1.Enabled := false;
      log('Disabled auto-check, refresh interval too low',false);
   end;
end;

procedure TfrmPageMonitor.WebBrowser1DocumentComplete(ASender: TObject;
  const pDisp: IDispatch; var URL: OleVariant);
var
   currentBrowser: IWebBrowser;
   topBrowser: IWebBrowser;
begin
   currentBrowser := pDisp as IWebBrowser;
   topBrowser := (ASender as TWebBrowser).DefaultInterface;
   if currentBrowser = topBrowser then
   begin
      // all frames have finished loading
      log('Page completely downloaded',false);
      startSearch;
      btnShowRawText.enabled := true;
   end
   else
      log('Page partly downloaded',true);
end;

procedure TfrmPageMonitor.WebBrowser1NavigateError(ASender: TObject;
  const pDisp: IDispatch; var URL, Frame, StatusCode: OleVariant;
  var Cancel: WordBool);
begin
   log('Navigate error',false);
end;

function TfrmPageMonitor.getPageContents: string;
var
   document: IHTMLDocument2;
begin
   document := WebBrowser1.Document as IHTMLDocument2;
   if Assigned(document) then
   begin
      // result := document.body.outerHTML; // gets the full HTML
      result := document.body.outerText; // gets the text with the HTML removed
   end
   else
      result := '';
end;

procedure TfrmPageMonitor.startSearch;
var
  searchResults: TStringList;
  contents: string;
  currentTimerState: boolean;
  i: integer;
  resp: TUserResponse;
begin
   log('Starting searching',true);
   contents := getPageContents;
   if length(contents) > 0 then
   begin
      searchResults := search(txtAlerts.Lines,contents);
      try
         log('Matches found: ' + IntToStr(searchResults.Count), false);
         if searchResults.Count > 0 then
         begin
            self.windowState := wsMaximized; // maximize this child window
            currentTimerState := Timer1.enabled; // suspend timer temporarily
            Timer1.enabled := false;
            try
               for i := 0 to searchResults.count-1 do
               begin
                  resp := (frmOwner as TMainForm).alertUser('Website match','Found match!', searchResults[i], WebBrowser1.LocationURL, false);
                  if resp.hadEnough then break;
               end;
            finally
               Timer1.enabled := currentTimerState; // reset timer
            end;
         end;
      finally
         searchResults.free;
      end;
   end;
   log('Finished searching',false);
end;

function TfrmPageMonitor.search(searchLines: TStrings; textToSearch: string): TStringList;
var
   l,i,j,found: integer;
   terms, webContentLines: TStringList;
   searchText: string;
begin
   result := TStringList.create;

   webContentLines := parseString(textToSearch, chr(13)+chr(10));
   try
      // check each line in searchLines
      for l := 0 to searchLines.count-1 do
      begin
         searchText := uppercase(searchLines[l]);

         // ignore searchLines that don't start with '1:'
         if copy(searchText, 1, 2) <> '1:' then
         begin
            log('Ignoring inactive keywords: ' + searchLines[l],true);
            continue;
         end;

         // remove '1:'
         searchText := copy(searchText, 3, length(searchText)-2);

         // ignore blank lines
         if length(searchText) = 0 then
         begin
            log('No keywords',true);
            continue;
         end;

         // first, break up searchText into individual terms
         // make sure searchText doesn't end with a comma
         terms := parseString(searchText, ',');
         try
            // now iterate over each line in the web page
            for j := 0 to webContentLines.count-1 do
            begin
               // look for the search terms in this line
               // if we find all the terms, we return the line they were found in
               found := 0;
               for i := 0 to terms.Count-1 do
               begin
                  // see if word must *not* be found on this line, e.g. !iran
                  if copy(terms[i], 1, 1) = '!' then
                  begin
                     if pos(copy(terms[i],2, length(terms[i])-1),uppercase(webContentLines[j])) > 0 then
                     begin
                        found := -1;
                        break;
                     end
                     else
                        inc(found);
                  end
                  else
                  begin
                     // see if word *must* be found on this line
                     if pos(terms[i],uppercase(webContentLines[j])) > 0 then inc(found);
                  end;
               end;
               if found = terms.count then
               begin
                  // check that an alert for this line hasn't been shown recently
                  if not lineShownRecently(webContentLines[j]) then
                  begin
                     log('Match found for ' + searchText + ' on line ' + webContentLines[j],false);
                     // add line to list of recently shown lines
                     // this means we won't show it again for a while
                     addLineToShownRecentlyList(webContentLines[j]);

                     result.add(webContentLines[j]);
                  end
                  else
                  begin
                     log('Match found but shown recently for ' + searchText + ' on line ' + webContentLines[j],false);
                  end;
               end;
            end;

         finally
            terms.free;
         end;
      end;
   finally
      webContentLines.free;
   end;
end;


procedure TfrmPageMonitor.setOwner(_frmOwner: TForm);
begin
   frmOwner := _frmOwner;
end;

procedure TfrmPageMonitor.loadConfig(prefix: string; var IniFile: TIniFile);
begin
   txtAddress.text := IniFile.readString(prefix, 'url', '');
   txtRefreshInterval.text := IniFile.readString(prefix, 'refreshInterval', '');
   txtAlerts.text := StringReplace(IniFile.readString(prefix, 'alerts', ''), 'CRLF', chr(13)+chr(10), [rfReplaceAll, rfIgnoreCase]);
   txtStartTime.text := IniFile.readString(prefix, 'startTimeToScan', '00:00');
   txtEndTime.text := IniFile.readString(prefix, 'endTimeToScan', '23:59');
end;


procedure TfrmPageMonitor.saveConfig(prefix: string; var IniFile: TIniFile);
begin
   IniFile.writeString(prefix, 'url',txtAddress.text);
   IniFile.writeString(prefix, 'refreshInterval',txtRefreshInterval.text);
   IniFile.writeString(prefix, 'alerts',StringReplace(txtAlerts.text, chr(13)+chr(10), 'CRLF', [rfReplaceAll, rfIgnoreCase]));
   IniFile.writeString(prefix, 'startTimeToScan',txtStartTime.text);
   IniFile.writeString(prefix, 'endTimeToScan',txtEndTime.text)
end;


procedure TfrmPageMonitor.addLineToShownRecentlyList(line: string);
begin
   linesShownRecently.put(line,FloatToStr(CurrentTimeMillis));
end;


function TfrmPageMonitor.lineShownRecently(line: string): boolean;
var
   i: integer;
begin
   result := false;

   for i := linesShownRecently.Count-1 downto 0 do
   begin
      // if line is old, remove it
      if CurrentTimeMillis - StrToFloat(linesShownRecently.getValueAt(i)) > repeatInterval then
      begin
         linesShownRecently.Delete(i);
      end
      else
      begin
         if line = linesShownRecently.getKeyAt(i) then
         begin
            result := true;
            exit;
         end;
      end;
   end;
end;

procedure TfrmPageMonitor.log(line: string; isVerbose: boolean);
begin
   if (not verboseOutput) and (isVerbose) then exit;
   if txtResults.Lines.Count > 10000 then txtResults.Clear;
   txtResults.Lines.Add(DateTimeToStr(Now) + ': ' + line);
   SendMessage(txtResults.handle, EM_SCROLLCARET,0,0);
end;

end.
