unit spider;

interface

uses
  math,shellapi,Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  strutils,IdHTTP,utils,Dialogs, StdCtrls, ExtCtrls, Grids,main,feeds, ComCtrls,
  showbigtext;

type
  TfrmSpider = class(TForm)
    pnlTop: TPanel;
    Label1: TLabel;
    txtStartingAddress: TEdit;
    btnStart: TButton;
    btnStop: TButton;
    pnlStatus: TPanel;
    Label4: TLabel;
    txtResults: TMemo;
    Splitter1: TSplitter;
    Label5: TLabel;
    lblCurrentPage: TLabel;
    lblShowRelevance: TLabel;
    tbShowRelevance: TTrackBar;
    lblCurrentPageDescription: TLabel;
    Label6: TLabel;
    cbSearchOffsiteLinks: TCheckBox;
    cbSearchHigherLinks: TCheckBox;
    cbShowLinks: TCheckBox;
    pnlFoundLinksAndItemsHolder: TPanel;
    pnlFoundPages: TPanel;
    sgPagesFound: TStringGrid;
    pnlFoundLinks: TPanel;
    sgLinksFound: TStringGrid;
    Splitter2: TSplitter;
    pnlFoundPagesHeader: TPanel;
    lblPagesFound: TLabel;
    btnOpenPage: TButton;
    btnPause: TButton;
    lblPagesChecked: TLabel;
    pnlFoundLinksHeader: TPanel;
    lblLinksFound: TLabel;
    btnOpenLink: TButton;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure setGui(enabled:boolean);
    procedure FormCreate(Sender: TObject);
    procedure btnStartClick(Sender: TObject);
    procedure btnStopClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure setColWidths;
    procedure FormDestroy(Sender: TObject);
    procedure Splitter2Moved(Sender: TObject);
    procedure btnOpenPageClick(Sender: TObject);
    procedure sgPagesFoundDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure cbShowLinksClick(Sender: TObject);
    procedure btnPauseClick(Sender: TObject);
    procedure btnOpenLinkClick(Sender: TObject);
  private
    { Private declarations }
    frmOwner: TForm;
    shouldRun,paused: boolean;
    linksFoundAlready: TStringList;
    procedure findLinksInPage(html: string);
    function getLinkContext(linkDescription, html: string; startOfLink, endOfLink: integer): string;
    function findNextLink: boolean;
    function linkDescriptionInLinksTable(description: string): boolean;
    procedure sortFoundLinks;
    procedure sortFoundPages;
    procedure pause;
    procedure log(line: string; isVerbose: boolean);
  public
    { Public declarations }
    procedure setOwner(_frmOwner: TForm);
  end;

const
   COL_LINK_URL = 0;
   COL_LINK_DESCRIPTION = 1;
   COL_LINK_RELEVANCE = 2;

   COL_PAGE_URL = 0;
   COL_PAGE_DESCRIPTION = 1;
   COL_PAGE_RELEVANCE = 2;
var
  frmSpider: TfrmSpider;

implementation


{$R *.dfm}

procedure TfrmSpider.setOwner(_frmOwner: TForm);
begin
   self.frmOwner := _frmOwner;
end;

procedure TfrmSpider.btnPauseClick(Sender: TObject);
begin
   paused := not paused;
   if paused then
   begin
      btnPause.Caption := 'Resume';
      screen.cursor := crDefault;
      log('Paused',false);
   end
   else
   begin
      btnPause.Caption := 'Pause';
      screen.cursor := crHourglass;
      log('Resumed',false);
   end;
   setGui(false);
end;

procedure TfrmSpider.btnStartClick(Sender: TObject);
var
   html,textWithoutHtml,pageText: string;
   pagesChecked,relevance: integer;
begin
   screen.Cursor := crHourglass;
   setGui(false);
   shouldRun := true;
   paused := false;
   linksFoundAlready.clear;
   pagesChecked := 0;
   sgLinksFound.RowCount := 1;
   sgPagesFound.RowCount := 1;
   lblCurrentPage.Caption := txtStartingAddress.text;
   linksFoundAlready.Add(txtStartingAddress.text);
   lblCurrentPageDescription.caption := '';
   frmFeeds.Timer1.Enabled := false;
   frmFeeds.cancelled := true;
   try
      while shouldRun do
      begin

         try
            try
               log('Downloading ' + lblCurrentPage.Caption,true);
               html := getWebPageAsString(lblCurrentPage.Caption, (frmOwner as TMainForm).proxyDetails);
               log('Downloaded ' + lblCurrentPage.Caption + ' ok',true);
            except
               on e:exception do
               begin
                  log('Error downloading ' + lblCurrentPage.Caption + ': ' + e.ClassName + ': ' + e.Message,false);
               end;
            end;

            processMessagesAtInterval;
            if not shouldRun then break;
            if paused then pause;

            if length(html) > 0 then
            begin
               // don't check relevance of initial page
               if lblCurrentPageDescription.caption <> '' then
               begin
                  log('Checking page relevance',false);
                  inc(pagesChecked);
                  lblPagesChecked.Caption := 'Pages checked:' + IntToStr(pagesChecked);
                  textWithoutHtml := removeHtml(html);
                  relevance := frmFeeds.checkRelevance(textWithoutHtml);
                  log('Page relevance=' + IntToStr(relevance) + '%',false);
                  processMessagesAtInterval;
                  if relevance > tbShowRelevance.position then
                  begin
                     with sgPagesFound do
                     begin
                        RowCount := RowCount+1;
                        RowHeights[RowCount-1] := 100;
                        FixedRows := 1;
                        cells[COL_PAGE_URL,RowCount-1] := lblCurrentPage.caption;
                        pageText := StringReplace(textWithoutHtml, chr(13), ' ', [rfReplaceAll, rfIgnoreCase]);
                        pageText := StringReplace(pageText, chr(10), ' ', [rfReplaceAll, rfIgnoreCase]);
                        pageText := StringReplace(pageText, chr(9), ' ', [rfReplaceAll, rfIgnoreCase]);
                        pageText := removeExcessSpaces(pageText);
                        cells[COL_PAGE_DESCRIPTION,RowCount-1] := lblCurrentPageDescription.caption + '- ' + copy(pageText,1,min(1000,length(pagetext)));
                        cells[COL_PAGE_RELEVANCE,RowCount-1] := IntToStr(relevance);
                     end;

                     processMessagesAtInterval;
                     if not shouldRun then break;
                     if paused then pause;

                     sortFoundPages;
                  end;
               end;

               processMessagesAtInterval;
               if not shouldRun then break;
               if paused then pause;

               findLinksInPage(html);
            end;

            processMessagesAtInterval;
            if not shouldRun then break;
            if paused then pause;

            if not findNextLink then
            begin
               log('Ran out of links!',false);
               shouldRun := false;
            end;

            lblLinksFound.Caption := 'Links found (' + IntToStr(sgLinksFound.RowCount-1) + ')';
            lblPagesFound.Caption := 'Pages found (' + IntToStr(sgPagesFound.RowCount-1) + ')';

            processMessagesAtInterval;
            if not shouldRun then break;
            if paused then pause;

         except
            on e:exception do
            begin
               log('Error: ' + e.ClassName + ': ' + e.Message,false);
            end;
         end;
      end;
   finally
      linksFoundAlready.clear;
      frmFeeds.Timer1.Enabled := true;
      btnPause.caption := 'Pause';
      Screen.Cursor := crDefault;
      setGui(true);
      log('Stopped searching',false);
      processMessagesAtInterval;
   end;
end;

procedure TfrmSpider.pause;
begin
   while paused do
   begin
      application.processmessages;
      sleep(500);
   end;
end;

procedure TfrmSpider.btnStopClick(Sender: TObject);
begin
   shouldRun := false;
   paused := false;
end;

procedure TfrmSpider.cbShowLinksClick(Sender: TObject);
begin
   pnlFoundLinks.Visible := cbShowLinks.Checked;
   splitter2.Visible := cbShowLinks.Checked;
   if cbShowLinks.Checked then
   begin
      with pnlFoundPages do
      begin
         visible := false;
         top := (pnlFoundLinksAndItemsHolder.height div 2)+1;
         left := 0;
         height := pnlFoundLinksAndItemsHolder.height div 2;
         width := pnlFoundLinksAndItemsHolder.width;
         align := alBottom;
      end;

      with pnlFoundLinks do
      begin
         Top := 0;
         Left := 0;
         Height := pnlFoundLinksAndItemsHolder.height div 2;
         Width := pnlFoundLinksAndItemsHolder.width;
         Align := alClient;
      end;

      pnlFoundPages.visible := true;
      setColWidths;
   end
   else
      pnlFoundPages.Align := alClient;
end;

procedure TfrmSpider.btnOpenLinkClick(Sender: TObject);
begin
   if sgLinksFound.Row > 0 then
   begin
      log('Opening ' + sgLinksFound.Cells[COL_LINK_URL,sgLinksFound.Row],false);
      ShellExecute(GetDesktopWindow, 'open', PChar(sgLinksFound.Cells[COL_LINK_URL,sgLinksFound.Row]), nil, nil, SW_ShowNormal);
   end
   else
      ShowMessage('Please select a link!');
end;

procedure TfrmSpider.btnOpenPageClick(Sender: TObject);
begin
   if sgPagesFound.Row > 0 then
   begin
      log('Opening ' + sgPagesFound.Cells[COL_PAGE_URL,sgPagesFound.Row],false);
      ShellExecute(GetDesktopWindow, 'open', PChar(sgPagesFound.Cells[COL_PAGE_URL,sgPagesFound.Row]), nil, nil, SW_ShowNormal);
   end
   else
      ShowMessage('Please select a page!');
end;

procedure TfrmSpider.FormClose(Sender: TObject; var Action: TCloseAction);
begin
   Action := caFree;
end;

procedure TfrmSpider.FormCreate(Sender: TObject);
begin
   randomize;
   linksFoundAlready := TStringList.create;
   linksFoundAlready.Sorted := true;
   setGui(true);
   sgLinksFound.Cells[COL_LINK_URL,0] := 'URL';
   sgLinksFound.Cells[COL_LINK_DESCRIPTION,0] := 'Description';
   sgLinksFound.Cells[COL_LINK_RELEVANCE,0] := 'Relevance (%)';

   sgPagesFound.Cells[COL_PAGE_URL,0] := 'URL';
   sgPagesFound.Cells[COL_PAGE_DESCRIPTION,0] := 'Text';
   sgPagesFound.Cells[COL_PAGE_RELEVANCE,0] := 'Relevance (%)';

   setColWidths;
   lblCurrentPageDescription.caption := '';

   lblShowRelevance.Hint := 'You can use this slider to indicate how relevant an web page needs to be before Webu decides to show it to you.' + chr(13) + chr(10) +
      'The program uses the feedback you have provided on the news feeds articles to estimate how relevant a particular page is.' + chr(13) + chr(10) + chr(13) + chr(10) +
      'If you move the slider towards LOW, you will get lots of web page that may have little relevance.' + chr(13) + chr(10) +
      'However, if you move the slider towards HIGH, Webu will only show you web page it thinks are very relevant.' + chr(13) + chr(10) +
      'Do not set it to either extreme though or you will get every single web page or none at all.' + chr(13) + chr(10) +
      'Try starting off by setting it somewhere in the middle and adjusting it slowly.';
   tbShowRelevance.Hint := lblShowRelevance.Hint;
   label6.Hint := lblShowRelevance.Hint;

   cbShowLinksClick(self);
end;

procedure TfrmSpider.FormDestroy(Sender: TObject);
begin
   linksFoundAlready.free;
end;

procedure TfrmSpider.FormResize(Sender: TObject);
begin
   setColWidths;
end;


procedure TfrmSpider.setColWidths;
begin
   sgLinksFound.colWidths[COL_LINK_URL] := (30*sgLinksFound.parent.Width) div 100;
   sgLinksFound.colWidths[COL_LINK_DESCRIPTION] := (60*sgLinksFound.parent.Width) div 100;
   sgLinksFound.colWidths[COL_LINK_RELEVANCE] := (10*sgLinksFound.parent.Width) div 100;

   sgPagesFound.colWidths[COL_PAGE_URL] := (30*sgPagesFound.parent.Width) div 100;
   sgPagesFound.colWidths[COL_PAGE_DESCRIPTION] := (60*sgPagesFound.parent.Width) div 100;
   sgPagesFound.colWidths[COL_PAGE_RELEVANCE] := (10*sgPagesFound.parent.Width) div 100;
end;

procedure TfrmSpider.setGui(enabled:boolean);
begin
   btnStart.Enabled := (enabled) and (not paused);
   btnStop.Enabled := (not enabled) and (not paused);
   btnPause.Enabled := (not enabled);
end;


procedure TfrmSpider.sgPagesFoundDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
var                       
   Sentence,                  { What is left in the cell to output }
   CurWord : String;          { The word we are currently outputting }
   SpacePos,                  { The position of the first space }
   CurX,                      { The x position of the 'cursor' }
   CurY : Integer;            { The y position of the 'cursor' }
   EndOfSentence : Boolean;   { Whether or not we are done outputting the cell }
begin
   // used code from http://www.bsdg.org/SWAG/DELPHI/0085.PAS.html

{ Initialize the font to be the control's font }
   Canvas.Font := Font;

   with sgPagesFound.Canvas do begin
      { If this is a fixed cell, then use the fixed color }
      // if gdFixed in State then begin
      if ARow = 0 then begin
         Pen.Color   := sgPagesFound.FixedColor;
         Brush.Color := sgPagesFound.FixedColor;
      end
      else if ARow = sgPagesFound.Row then
      begin
         Pen.Color   := sgPagesFound.Color;
         Brush.Color := clAqua;
      end
      { else, use the normal color }
      else begin
         Pen.Color   := sgPagesFound.Color;
         Brush.Color := sgPagesFound.Color;
      end;

      { Prepaint cell in cell color }
      Rectangle(Rect.Left, Rect.Top, Rect.Right, Rect.Bottom);
   end;

   { Start the drawing in the upper left corner of the cell }
   CurX := Rect.Left;
   CurY := Rect.Top;

   { Here we get the contents of the cell }
   Sentence := sgPagesFound.Cells[ACol, ARow];

   { for each word in the cell }
   EndOfSentence := FALSE;
   while (not EndOfSentence) do begin
      { to get the next word, we search for a space }
      SpacePos := Pos(' ', Sentence);
      if SpacePos > 0 then begin
         { get the current word plus the space }
         CurWord := Copy(Sentence, 0, SpacePos);

         { get the rest of the sentence }
         Sentence := Copy(Sentence, SpacePos + 1, Length(Sentence) - SpacePos);
      end
      else begin
         { this is the last word in the sentence }
         EndOfSentence := TRUE;
         CurWord := Sentence;
      end;

      with sgPagesFound.Canvas do begin
         { if the text goes outside the boundary of the cell }
         if (TextWidth(CurWord) + CurX) > Rect.Right then begin
            { wrap to the next line }
            CurY := CurY + TextHeight(CurWord);
            CurX := Rect.Left;
         end;

         { write out the word }
         TextOut(CurX, CurY, CurWord);
         { increment the x position of the cursor }
         CurX := CurX + TextWidth(CurWord);
      end;
   end;
end;


procedure TfrmSpider.log(line: string; isVerbose: boolean);
begin
   if (not verboseOutput) and (isVerbose) then exit;
   if txtResults.Lines.Count > 10000 then txtResults.Clear;
   txtResults.Lines.Add(DateTimeToStr(Now) + ': ' + line);
   SendMessage(txtResults.handle, EM_SCROLLCARET,0,0);
end;

procedure TfrmSpider.findLinksInPage(html: string);
var
   currentPos,startOfLink,endOfUrl,startOfText,endOfLink: integer;
   uppercaseHTML, link, linkDescription: string;
begin
   log('Searching page for hyperlinks',false);
   currentPos := 1;
   startOfText := 0;
   endOfLink := 0;
   html := removeExcessSpaces(html);
   //showBigTextForm(html);
   uppercaseHTML := uppercase(html);
   repeat
      processMessagesAtInterval;
      startOfLink := posex('<A HREF="', uppercaseHTML, currentPos);
      if startOfLink > 0 then
      begin
         currentPos := startOfLink+1;
         endOfUrl := posex('"',uppercaseHTML,startOfLink+10);
         startOfText := posex('>',uppercaseHTML,endOfUrl);
         if (endOfUrl > 0) and (startOfText > 0) then
         begin
            link := copy(html,startOfLink+9,endOfUrl - (startOfLink+9));
            endOfLink := posex('</A>',uppercaseHTML,startOfText);
            if endOfLink > 0 then
            begin
               with sgLinksFound do
               begin
                  link := StringReplace(link, '"', '', [rfReplaceAll, rfIgnoreCase]);
                  if copy(link,1,1) = '#' then continue;
                  if pos('HTTPS:',uppercase(link)) > 0 then continue;
                  if pos('JAVASCRIPT',uppercase(link)) > 0 then continue;
                  if pos('VBSCRIPT',uppercase(link)) > 0 then continue;
                  if copy(link,1,2) = '//' then link := 'http:' + link;
                  if copy(link,1,1) = '/' then link := getBaseUrl(lblCurrentPage.Caption) + link;
                  if uppercase(copy(link,1,7)) <> 'HTTP://' then link := getBaseDir(lblCurrentPage.Caption) + '/' + link;
                  if (not cbSearchOffsiteLinks.Checked) and (getBaseUrl(lblCurrentPage.Caption) <> getBaseUrl(link)) then continue;
                  if (not cbSearchHigherLinks.Checked) and
                     (pos(lblCurrentPage.Caption, link) = 0) then continue;

                  linkDescription := copy(html,startOfText+1,endOfLink-startOfText-1);
                  linkDescription := removeHtml(linkDescription);

                  // take the link description, plus the 50 characters either side of it (if any)
                  linkDescription := getLinkContext(linkDescription, html, startOfLink, endOfLink+4);

                  if (not findStringInSortedList(link,linksFoundAlready)) and (not linkDescriptionInLinksTable(linkDescription)) then
                  begin
                     RowCount := RowCount + 1;
                     FixedRows := 1;
                     cells[COL_LINK_URL,rowCount-1] := link;
                     cells[COL_LINK_DESCRIPTION,rowCount-1] := linkDescription;
                     cells[COL_LINK_RELEVANCE,rowCount-1] := IntToStr(frmFeeds.checkRelevance(linkDescription));
                     linksFoundAlready.Add(link);
                  end;
               end;
            end;
         end;
      end;
   until (startOfLink = 0) or (startOfText = 0) or (endOfLink = 0);
end;

function TfrmSpider.getLinkContext(linkDescription, html: string; startOfLink, endOfLink: integer): string;
const
   maxExtra = 50;
var
   lastSpacePos: integer;
   prefix,suffix: string;
   startingPos: integer;
begin
   // get prefix
   lastSpacePos := -1;
   startingPos := startOfLink;
   while (startingPos > 1) and (startOfLink - startingPos < maxExtra) do
   begin
      if html[startingPos] = ' ' then lastSpacePos := startingPos;
      dec(startingPos);
   end;
   if lastSpacePos > -1 then
      prefix := removeHtml(copy(html, lastSpacePos, startOfLink-lastSpacePos-1))
   else
      prefix := '';

   // get suffix
   lastSpacePos := -1;
   startingPos := endOfLink;
   while (startingPos < length(html)) and (startingPos - endOfLink < maxExtra) do
   begin
      if html[startingPos] = ' ' then lastSpacePos := startingPos;
      inc(startingPos);
   end;
   if lastSpacePos > -1 then
      suffix := removeHtml(copy(html, endOfLink, lastSpacePos-endOfLink))
   else
      suffix := '';

   result := trim(prefix + ' ' + linkDescription + ' ' + suffix);
end;

function TfrmSpider.findNextLink: boolean;
var
   index: integer;
begin
   result := sgLinksFound.RowCount > 1;
   log('Sorting links',false);
   sortFoundLinks;
   if result then
   begin
      // occasionally pick a random link just to stop it getting stuck
      // going 100% after a single story
      if (sgLinksFound.RowCount > 5) and (random(10) = 5) then
      begin
         index := random(sgLinksFound.RowCount-1)+1;
      end
      else
         index := 1;
      lblCurrentPage.Caption := sgLinksFound.Cells[COL_LINK_URL, index];
      log('Next link is ' + lblCurrentPage.Caption,false);
      lblCurrentPageDescription.caption := sgLinksFound.Cells[COL_LINK_DESCRIPTION, 1];
      lblCurrentPageDescription.Left := lblCurrentPage.left + lblCurrentPage.width + 20;
      deleteRow(sgLinksFound,index);
   end;
end;

procedure TfrmSpider.sortFoundLinks;
var
   i,j,c: integer;
   tmp: string;
begin
   for i := 1 to sgLinksFound.RowCount - 1 do
   begin
      for j := 1 to sgLinksFound.RowCount - 1 do
      begin
         if StrToInt(sgLinksFound.cells[COL_LINK_RELEVANCE,i]) >  StrToInt(sgLinksFound.cells[COL_LINK_RELEVANCE,j]) then
         begin
            for c := 0 to sgLinksFound.colCount-1 do
            begin
               tmp := sgLinksFound.cells[c,i];
               sgLinksFound.cells[c,i] := sgLinksFound.cells[c,j];
               sgLinksFound.cells[c,j] := tmp;
            end;
         end;
      end;
   end;
end;

procedure TfrmSpider.sortFoundPages;
var
   i,j,c: integer;
   tmp: string;
begin
   log('Sorting found pages',false);
   for i := 1 to sgPagesFound.RowCount - 1 do
   begin
      for j := 1 to sgPagesFound.RowCount - 1 do
      begin
         if StrToInt(sgPagesFound.cells[COL_PAGE_RELEVANCE,i]) >  StrToInt(sgPagesFound.cells[COL_PAGE_RELEVANCE,j]) then
         begin
            for c := 0 to sgPagesFound.colCount-1 do
            begin
               tmp := sgPagesFound.cells[c,i];
               sgPagesFound.cells[c,i] := sgPagesFound.cells[c,j];
               sgPagesFound.cells[c,j] := tmp;
            end;
         end;
      end;
   end;
end;


procedure TfrmSpider.Splitter2Moved(Sender: TObject);
begin
   setColWidths;
end;


function TfrmSpider.linkDescriptionInLinksTable(description: string): boolean;
var
   i: integer;
begin
   result := false;
   for i := 1 to sgLinksFound.rowCount-1 do
   begin
      if sgLinksFound.cells[COL_LINK_DESCRIPTION,i] = description then
      begin
         result := true;
         exit;
      end;
   end;
end;


end.
